__all__ = ('AllowedMentionProxy',)

from scarletio import RichAttributeErrorBaseType, export

from ..role import Role, create_partial_role_from_id
from ..user import UserBase, create_partial_user_from_id

from .constants import STATE_ALLOW_REPLIED_USER_FALSE, STATE_ALLOW_REPLIED_USER_NONE, STATE_ALLOW_REPLIED_USER_TRUE

from .helpers import (
    _nullable_list_difference, _nullable_list_intersection, _nullable_list_symmetric_difference, _nullable_list_union
)
from .utils import is_allowed_mentions_valid


@export
class AllowedMentionProxy(RichAttributeErrorBaseType):
    """
    Proxy class to interact with allowed mentions.
    
    Attributes
    ----------
    _allow_everyone : `int`
        Whether everyone mention is enabled.
    _allow_replied_user : `int`
        Whether replied user can be mentioned.
    _allow_roles : `int`
        Whether all role is enabled.
    _allow_users : `int`
        Whether all user is enabled.
    _allowed_role_ids : `None | list<int>`
        The enabled roles by the proxy.
    _allowed_user_ids : `None | list<int>
        The enabled users by the proxy.
    """
    __slots__ = (
        '_allow_everyone', '_allow_replied_user', '_allow_roles', '_allow_users', '_allowed_role_ids', '_allowed_user_ids'
    )
    
    def __new__(cls, *allowed_mentions):
        """
        Parses allowed mentions
        
        Parameters
        ----------
        *allowed_mentions : `str`, ``UserBase``, ``Role``
            Which user or role can the message ping (or everyone).
        
        Raises
        ------
        TypeError
            If `allowed_mentions` contains an element of invalid type.
        ValueError
            If `allowed_mentions` contains en element of correct type, but an invalid value.
        """
        allow_users = 0
        allow_roles = 0
        allow_everyone = 0
        allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        allowed_role_ids = None
        allowed_user_ids = None
        
        if allowed_mentions:
            for element in allowed_mentions:
                if isinstance(element, str):
                    if element == '!replied_user':
                        allow_replied_user = STATE_ALLOW_REPLIED_USER_FALSE
                        continue
                    
                    if element == 'replied_user':
                        allow_replied_user = STATE_ALLOW_REPLIED_USER_TRUE
                        continue
                    
                    if element == 'everyone':
                        allow_everyone = 1
                        continue
                    
                    if element == 'users':
                        allow_users = 1
                        continue
                    
                    if element == 'roles':
                        allow_roles = 1
                        continue
                    
                    raise ValueError(
                        f'`allowed_mentions` contains a not valid `str` element: `{element!r}`. `str` '
                        f'elements can be any of: (\'everyone\', \'users\', \'roles\', \'replied_user\', '
                        f'\'!replied_user\').'
                    )
                
                if isinstance(element, UserBase):
                    if allowed_user_ids is None:
                        allowed_user_ids = []
                    
                    allowed_user_ids.append(element.id)
                    continue
                
                if isinstance(element, Role):
                    if allowed_role_ids is None:
                        allowed_role_ids = []
                    
                    allowed_role_ids.append(element.id)
                    continue
                
                raise TypeError(
                    f'`allowed_mentions` can contain `str`, `{Role.__name__}`, `{UserBase.__name__}` elements, got '
                    f' {type(element).__name__}; {element!r}; allowed_mentions = {allowed_mentions!r}.'
                )
            
            if allow_users:
                allowed_user_ids = None
            
            if allow_roles:
                allowed_role_ids = None
        
            
        self = object.__new__(cls)
        self._allow_users = allow_users
        self._allow_roles = allow_roles
        self._allow_everyone = allow_everyone
        self._allow_replied_user = allow_replied_user
        self._allowed_role_ids = allowed_role_ids
        self._allowed_user_ids = allowed_user_ids
        return self
    
    
    @classmethod
    def _create_from_various(cls, other):
        """
        Creates an allowed mention proxy instance from other object.
        
        Parameters
        ----------
        other : `instance<cls> | str, (list | set | tuple)<str | UserBase, Role>`
            The other instance to create allowed mention proxy from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if isinstance(other, cls):
            return other
        
        if isinstance(other, (list, set, tuple)):
            return cls(*other)
        
        return cls(other)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new allowed mention proxy from the given data.
        
        Parameters
        ----------
        data : `None | dict<str, object>`items
            Allowed mention data.
        
        Returns
        -------
        self : `instance<cls>`
            The created allowed mention proxy.
        """
        if (data is None) or (not data):
            allow_users = 0
            allow_roles = 0
            allow_everyone = 0
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
            allowed_role_ids = None
            allowed_user_ids = None
            
        else:
            try:
                allow_replied_user_raw = data['replied_user']
            except KeyError:
                allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
            else:
                if allow_replied_user_raw:
                    allow_replied_user = STATE_ALLOW_REPLIED_USER_TRUE
                else:
                    allow_replied_user = STATE_ALLOW_REPLIED_USER_FALSE
            
            allowed_role_ids_raw = data.get('roles', None)
            if (allowed_role_ids_raw is None) or (not allowed_role_ids_raw):
                allowed_role_ids = None
            else:
                allowed_role_ids = []
                for role_id in allowed_role_ids_raw:
                    role_id = int(role_id)
                    allowed_role_ids.append(role_id)
            
            allowed_user_ids_raw = data.get('users', None)
            if (allowed_user_ids_raw is None) or (not allowed_user_ids_raw):
                allowed_user_ids = None
            else:
                allowed_user_ids = []
                for user_id in allowed_user_ids_raw:
                    user_id = int(user_id)
                    allowed_user_ids.append(user_id)
            
            try:
                parse_all_of = data['parse']
            except KeyError:
                allow_users = 0
                allow_roles = 0
                allow_everyone = 0
            else:
                if 'everyone' in parse_all_of:
                    allow_everyone = 1
                else:
                    allow_everyone = 0
                
                if 'roles' in parse_all_of:
                    allow_roles = 1
                else:
                    allow_roles = 0
                
                if 'users' in parse_all_of:
                    allow_users = 1
                else:
                    allow_users = 0
        
        self = object.__new__(cls)
        self._allow_users = allow_users
        self._allow_roles = allow_roles
        self._allow_everyone = allow_everyone
        self._allow_replied_user = allow_replied_user
        self._allowed_role_ids = allowed_role_ids
        self._allowed_user_ids = allowed_user_ids
        return self
    
    
    def to_data(self):
        """
        Converts the allowed mention proxy to json serializable object.
        
        Returns
        -------
        data : `dict<str, object> items
        """
        data = {}
        parse_all_of = None
        
        allow_replied_user = self._allow_replied_user
        if allow_replied_user != STATE_ALLOW_REPLIED_USER_NONE:
            data['replied_user'] = (allow_replied_user == STATE_ALLOW_REPLIED_USER_TRUE)
        
        if self._allow_everyone:
            if parse_all_of is None:
                parse_all_of = []
                data['parse'] = parse_all_of
            
            parse_all_of.append('everyone')
        
        if self._allow_users:
            if parse_all_of is None:
                parse_all_of = []
                data['parse'] = parse_all_of
            
            parse_all_of.append('users')
        else:
            allowed_user_ids = self._allowed_user_ids
            if (allowed_user_ids is not None):
                data['users'] = [str(user_id) for user_id in allowed_user_ids]
        
        if self._allow_roles:
            if parse_all_of is None:
                parse_all_of = []
                data['parse'] = parse_all_of
            
            parse_all_of.append('roles')
        else:
            allowed_role_ids = self._allowed_role_ids
            if (allowed_role_ids is not None):
                data['roles'] = [str(role_id) for role_id in allowed_role_ids]
        
        # Set nothing to parse if we did not set anything.
        if not data:
            data['parse'] = []
        
        return data
    
    
    def __repr__(self):
        """Returns the allowed mention proxy's representation."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # allow_everyone
        allow_everyone = self._allow_everyone
        if allow_everyone:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allow_everyone = ')
            repr_parts.append(repr(True if allow_everyone else False))
        
        # allow_replied_user
        allow_replied_user = self._allow_replied_user
        if allow_replied_user != STATE_ALLOW_REPLIED_USER_NONE:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allow_replied_user = ')
            repr_parts.append(repr((allow_replied_user == STATE_ALLOW_REPLIED_USER_TRUE)))
        
        # allow_roles
        allow_roles = self._allow_roles
        if allow_roles:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allow_roles = ')
            repr_parts.append(repr(True if allow_roles else allow_roles))
        
        # allow_users
        allow_users = self._allow_users
        if allow_users:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allow_users = ')
            repr_parts.append(repr(True if allow_users else False))
        
        # allowed_role_ids
        allowed_role_ids = self._allowed_role_ids
        if (allowed_role_ids is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allowed_role_ids = [')
            
            limit = len(allowed_role_ids)
            index = 0
            
            while True:
                role_id = allowed_role_ids[index]
                repr_parts.append(repr(role_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # allowed_user_ids
        allowed_user_ids = self._allowed_user_ids
        if (allowed_user_ids is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' allowed_user_ids = [')
            
            limit = len(allowed_user_ids)
            index = 0
            
            while True:
                user_id = allowed_user_ids[index]
                repr_parts.append(repr(user_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the allowed mention proxy.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        allowed_role_ids = self._allowed_role_ids
        if (allowed_role_ids is not None):
            allowed_role_ids = allowed_role_ids.copy()
        
        allowed_user_ids = self._allowed_user_ids
        if (allowed_user_ids is not None):
            allowed_user_ids = allowed_user_ids.copy()
        
        new = object.__new__(type(self))
        new._allow_everyone = self._allow_everyone
        new._allow_replied_user = self._allow_replied_user
        new._allow_roles = self._allow_roles
        new._allow_users = self._allow_users
        new._allowed_role_ids = allowed_role_ids
        new._allowed_user_ids = allowed_user_ids
        return new
    
    
    def __eq__(self, other):
        """Returns whether the two allowed mention proxies are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self._allow_users != other._allow_users:
            return False
        
        if self._allow_roles != other._allow_roles:
            return False
        
        if self._allow_everyone != other._allow_everyone:
            return False
        
        if self._allow_replied_user != other._allow_replied_user:
            return False
        
        if self._allowed_role_ids != other._allowed_role_ids:
            return False
        
        if self._allowed_user_ids != other._allowed_user_ids:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the allowed mention proxy's hash value."""
        hash_value = 0
        hash_value ^= self._allow_users << 3
        hash_value ^= self._allow_roles << 6
        hash_value ^= self._allow_everyone << 9
        hash_value ^= self._allow_replied_user << 12
        
        allowed_role_ids = self._allowed_role_ids
        if (allowed_role_ids is not None):
            hash_value ^= len(allowed_role_ids) << 16
            for role_id in allowed_role_ids:
                hash_value ^= role_id

        allowed_user_ids = self._allowed_user_ids
        if (allowed_user_ids is not None):
            hash_value ^= len(allowed_user_ids) << 24
            for user_id in allowed_user_ids:
                hash_value ^= user_id
        
        return hash_value
    
    
    def __and__(self, other):
        """Returns the intersection of the two allowed mention proxy."""
        if not is_allowed_mentions_valid(other):
            return NotImplemented
        
        other = self._create_from_various(other)
        
        
        allow_roles = self._allow_roles & other._allow_roles
        
        allow_users = self._allow_users & other._allow_users
        
        allow_everyone = self._allow_everyone & other._allow_everyone
        
        allow_replied_user = self._allow_replied_user
        if allow_replied_user != other._allow_replied_user:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        
        if allow_roles:
            allowed_role_ids = None
        
        elif self._allow_roles:
            allowed_role_ids = other._allowed_role_ids
            if (allowed_role_ids is not None):
                allowed_role_ids = allowed_role_ids.copy()
        
        elif other._allow_roles:
            allowed_role_ids = self._allowed_role_ids
            if (allowed_role_ids is not None):
                allowed_role_ids = allowed_role_ids.copy()
        else:
            allowed_role_ids = _nullable_list_intersection(self._allowed_role_ids, other._allowed_role_ids)
        
        if allow_users:
            allowed_user_ids = None
        
        elif self._allow_users:
            allowed_user_ids = other._allowed_user_ids
            if (allowed_user_ids is not None):
                allowed_user_ids = allowed_user_ids.copy()
        
        elif other._allow_users:
            allowed_user_ids = self._allowed_user_ids
            if (allowed_user_ids is not None):
                allowed_user_ids = allowed_user_ids.copy()
        else:
            allowed_user_ids = _nullable_list_intersection(self._allowed_user_ids, other._allowed_user_ids)
        
        new = object.__new__(type(self))
        new._allow_everyone = allow_everyone
        new._allow_replied_user = allow_replied_user
        new._allow_roles = allow_roles
        new._allow_users = allow_users
        new._allowed_role_ids = allowed_role_ids
        new._allowed_user_ids = allowed_user_ids
        
        return new
    
    
    __rand__ = __and__
    
    
    def __xor__(self, other):
        """Returns the symmetric difference of the two allowed mention proxy."""
        if not is_allowed_mentions_valid(other):
            return NotImplemented
        
        other = self._create_from_various(other)
        
        
        allow_roles = self._allow_roles ^ other._allow_roles
        
        allow_users = self._allow_users ^ other._allow_users
        
        allow_everyone = self._allow_everyone ^ other._allow_everyone
        
        self_allow_replied_user = self._allow_replied_user
        other_allow_replied_user = other._allow_replied_user
        if self_allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user = other_allow_replied_user
        elif other_allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user = self_allow_replied_user
        else:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        
        if allow_roles:
            allowed_role_ids = None
        else:
            allowed_role_ids = _nullable_list_symmetric_difference(self._allowed_role_ids, other._allowed_role_ids)
        
        if allow_users:
            allowed_user_ids = None
        else:
            allowed_user_ids = _nullable_list_symmetric_difference(self._allowed_user_ids, other._allowed_user_ids)
        
        new = object.__new__(type(self))
        new._allow_everyone = allow_everyone
        new._allow_replied_user = allow_replied_user
        new._allow_roles = allow_roles
        new._allow_users = allow_users
        new._allowed_role_ids = allowed_role_ids
        new._allowed_user_ids = allowed_user_ids
        
        return new
    
    
    __rxor__ = __xor__
    
    
    def __or__(self, other):
        """Returns the union of the two allowed mention proxy."""
        if not is_allowed_mentions_valid(other):
            return NotImplemented
        
        other = self._create_from_various(other)
        
        allow_roles = self._allow_roles | other._allow_roles
        
        allow_users = self._allow_users | other._allow_users
        
        allow_everyone = self._allow_everyone | other._allow_everyone
        
        self_allow_replied_user = self._allow_replied_user
        other_allow_replied_user = other._allow_replied_user
        if self_allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user = other_allow_replied_user
        elif other_allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user = self_allow_replied_user
        elif self_allow_replied_user == other_allow_replied_user:
            allow_replied_user = self_allow_replied_user
        else:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        
        if allow_roles:
            allowed_role_ids = None
        else:
            allowed_role_ids = _nullable_list_union(self._allowed_role_ids, other._allowed_role_ids)
        
        if allow_users:
            allowed_user_ids = None
        else:
            allowed_user_ids = _nullable_list_union(self._allowed_user_ids, other._allowed_user_ids)
        
        
        new = object.__new__(type(self))
        new._allow_everyone = allow_everyone
        new._allow_replied_user = allow_replied_user
        new._allow_roles = allow_roles
        new._allow_users = allow_users
        new._allowed_role_ids = allowed_role_ids
        new._allowed_user_ids = allowed_user_ids
        
        return new
    
    
    __ror__ = __or__
    __add__ = __or__
    __radd__ = __or__
    
    
    @classmethod
    def _subtract(cls, self, other):
        """
        Returns an allowed mentions proxy, without elements found in other.
        
        This is a classmethod.
        
        Parameters
        ----------
        self : `instance<type<self>>`
            The allowed mention proxy to subtract the other from.
        other : `instance<type<self>>`
            The allowed mention proxy to subtract.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if other._allow_roles:
            allow_roles = 0
        else:
            allow_roles = self._allow_roles
        
        if other._allow_users:
            allow_users = 0
        else:
            allow_users = self._allow_users
        
        if other._allow_everyone:
            allow_everyone = 0
        else:
            allow_everyone = self._allow_everyone
        
        self_allow_replied_user = self._allow_replied_user
        other_allow_replied_user = other._allow_replied_user
        if self_allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user = self_allow_replied_user
        elif other_allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user = self_allow_replied_user
        elif self_allow_replied_user == other_allow_replied_user:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        else:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        
        if allow_roles or other._allow_roles:
            allowed_role_ids = None
        else:
            allowed_role_ids = _nullable_list_difference(self._allowed_role_ids, other._allowed_role_ids)
        
        if allow_users or other._allow_users:
            allowed_user_ids = None
        else:
            allowed_user_ids = _nullable_list_difference(self._allowed_user_ids, other._allowed_user_ids)
        
        new = object.__new__(type(self))
        new._allow_roles = allow_roles
        new._allow_users = allow_users
        new._allow_everyone = allow_everyone
        new._allow_replied_user = allow_replied_user
        new._allowed_role_ids = allowed_role_ids
        new._allowed_user_ids = allowed_user_ids
        
        return new
    
    
    def __sub__(self, other):
        """Returns an allowed mentions proxy, without elements found in other."""
        if not is_allowed_mentions_valid(other):
            return NotImplemented
        
        other = self._create_from_various(other)
        return self._subtract(self, other)
    
    
    def __rsub__(self, other):
        """Returns an allowed mentions proxy, without elements found in self."""
        if not is_allowed_mentions_valid(other):
            return NotImplemented
        
        other = self._create_from_various(other)
        return self._subtract(other, self)
    
    
    def __iter__(self):
        """
        Iterates back the values which with a same allowed mention proxy can be created.
        
        This method is an iterable generator.
        
        Yields
        ------
        allowed_mentions : `str`, ``UserBase``, ``Role``
            Which user or role can the message ping (or everyone).
        """
        
        if self._allow_everyone:
            yield 'everyone'
        
        allow_replied_user = self._allow_replied_user
        if allow_replied_user == STATE_ALLOW_REPLIED_USER_TRUE:
            yield 'replied_user'
        elif allow_replied_user == STATE_ALLOW_REPLIED_USER_FALSE:
            yield '!replied_user'
        
        if self._allow_roles:
            yield 'roles'
        
        if self._allow_users:
            yield 'users'
        
        
        allowed_role_ids = self._allowed_role_ids
        if (allowed_role_ids is not None):
            for role_id in allowed_role_ids:
                yield create_partial_role_from_id(role_id)
        
        allowed_user_ids = self._allowed_user_ids
        if (allowed_user_ids is not None):
            for user_id in allowed_user_ids:
                yield create_partial_user_from_id(user_id)
    
    
    def update(self, other):
        """
        Updates the allowed mentions with the given value.
        
        Parameters
        ----------
        other : `str`, ``UserBase``, ``Role``, `instance<type<self>>` or (`list`, `tuple`, `set`) of \
                (`str`, ``UserBase``, ``Role``)
            Which user or role can the message ping (or everyone).
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            If `other` contains an element of invalid type.
        ValueError
            If `other` contains en element of correct type, but an invalid value.
        """
        other = self._create_from_various(other)
        
        allow_replied_user = other._allow_replied_user
        if allow_replied_user != STATE_ALLOW_REPLIED_USER_NONE:
            self._allow_replied_user = allow_replied_user
        
        allow_everyone = other._allow_everyone
        if allow_everyone:
            self._allow_everyone = allow_everyone
        
        allow_roles = other._allow_roles
        if allow_roles:
            self._allow_roles = allow_roles
            self._allowed_role_ids = None
        else:
            if not self._allow_roles:
                self._allowed_role_ids = _nullable_list_union(self._allowed_role_ids, other._allowed_role_ids)
        
        
        allow_users = other._allow_users
        if allow_users:
            self._allow_users = allow_users
            self._allowed_user_ids = None
        else:
            if not self._allow_users:
                self._allowed_user_ids = _nullable_list_union(self._allowed_user_ids, other._allowed_user_ids)
    
    
    @property
    def allow_roles(self):
        """
        A get-set-del property to enable or disable role mentions.
        
        Accepts and returns `bool`-s.
        """
        if self._allow_roles:
            allow_roles_output = True
        else:
            allow_roles_output = False
        
        return allow_roles_output
    
    
    @allow_roles.setter
    def allow_roles(self, allow_roles_input):
        if not isinstance(allow_roles_input, bool):
            raise TypeError(
                f'`allow_roles` can be `bool`, got '
                f'{type(allow_roles_input).__name__}; {allow_roles_input!r}.'
            )
        
        if allow_roles_input:
            allow_roles = 1
            self._allowed_role_ids = None
        else:
            allow_roles = 0
        
        self._allow_roles = allow_roles
    
    
    @allow_roles.deleter
    def allow_roles(self):
        self._allow_roles = 0
    
    
    @property
    def allow_users(self):
        """
        A get-set-del property to enable or disable user mentions.
        
        Accepts and returns `bool`-s.
        """
        if self._allow_users:
            allow_users_output = True
        else:
            allow_users_output = False
        
        return allow_users_output
    
    
    @allow_users.setter
    def allow_users(self, allow_users_input):
        if not isinstance(allow_users_input, bool):
            raise TypeError(
                f'`allow_users` can be `bool`, got '
                f'{type(allow_users_input).__name__}; {allow_users_input!r}.'
            )
        
        if allow_users_input:
            allow_users = 1
            self._allowed_user_ids = None
        else:
            allow_users = 0
        
        self._allow_users = allow_users
    
    
    @allow_users.deleter
    def allow_users(self):
        self._allow_users = 0
    
    
    @property
    def allow_everyone(self):
        """
        A get-set-del property to enable or disable everyone mentions.
        
        Accepts and returns `bool`-s.
        """
        if self._allow_everyone:
            allow_everyone_output = True
        else:
            allow_everyone_output = False
        
        return allow_everyone_output
    
    
    @allow_everyone.setter
    def allow_everyone(self, allow_everyone_input):
        if not isinstance(allow_everyone_input, bool):
            raise TypeError(
                f'`allow_everyone` can be `bool`, got '
                f'{type(allow_everyone_input).__name__}; {allow_everyone_input!r}.'
            )
        
        if allow_everyone_input:
            allow_everyone = 1
        else:
            allow_everyone = 0
        
        self._allow_everyone = allow_everyone
    
    
    @allow_everyone.deleter
    def allow_everyone(self):
        self._allow_everyone = 0
    
    
    @property
    def allow_replied_user(self):
        """
        A get-set-del property to enable or disable replied user mention.
        
        Accepts and returns `None` and `bool`-s.
        """
        allow_replied_user = self._allow_replied_user
        if allow_replied_user == STATE_ALLOW_REPLIED_USER_NONE:
            allow_replied_user_output = None
        elif allow_replied_user == STATE_ALLOW_REPLIED_USER_TRUE:
            allow_replied_user_output = True
        else:
            allow_replied_user_output = False
        
        return allow_replied_user_output
    
    
    @allow_replied_user.setter
    def allow_replied_user(self, allow_replied_user_input):
        if (allow_replied_user_input is not None) and (not isinstance(allow_replied_user_input, bool)):
            raise TypeError(
                f'`allow_replied_user` can be `None`, `int`, got '
                f'{type(allow_replied_user_input).__name__}; {allow_replied_user_input!r}.'
            )
        
        if allow_replied_user_input is None:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
        elif allow_replied_user_input:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_TRUE
        else:
            allow_replied_user = STATE_ALLOW_REPLIED_USER_FALSE
        
        self._allow_replied_user = allow_replied_user
    
    
    @allow_replied_user.deleter
    def allow_replied_user(self):
        self._allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
    
    
    @property
    def allowed_roles(self):
        """
        A get-set-del property to enable or disable specific role mentions.
        
        Accepts and returns `None` or a `list`, `set` or `tuple` of ``Role``-s.
        """
        allowed_role_ids = self._allowed_role_ids
        if allowed_role_ids is not None:
            return [create_partial_role_from_id(role_id) for role_id in allowed_role_ids]
    
    
    @allowed_roles.setter
    def allowed_roles(self, allowed_roles):
        if self._allow_roles:
            return
        
        if allowed_roles is None:
            allowed_role_ids = None
        
        elif isinstance(allowed_roles, Role):
            allowed_role_ids = [allowed_roles.id]
        
        elif isinstance(allowed_roles, (list, set, tuple)):
            allowed_role_ids = None
            
            for role in allowed_roles:
                if not isinstance(role, Role):
                    raise TypeError(
                        f'`allowed_roles` can contain `{Role.__name__}` elements, got '
                        f'{type(role).__name__}; {role!r}; allowed_role_ids_input = {allowed_roles!r}.'
                    )
                
                if allowed_role_ids is None:
                    allowed_role_ids = []
                
                allowed_role_ids.append(role.id)
        else:
            raise TypeError(
                f'`allowed_roles` can be `None`, `{Role.__name__}`, `list`, `tuple`, `set` of {Role.__name__}`'
                f', got {type(allowed_roles).__name__}; {allowed_roles!r}.'
            )
        
        self._allowed_role_ids = allowed_role_ids
    
    
    @allowed_roles.deleter
    def allowed_roles(self):
        self._allowed_role_ids = None
    

    @property
    def allowed_users(self):
        """
        A get-set-del property to enable or disable specific user mentions.
        
        Accepts and returns `None` or a `list`, `set` or `tuple` of ``UserBase``-s.
        """
        allowed_user_ids = self._allowed_user_ids
        if allowed_user_ids is not None:
            return [create_partial_user_from_id(user_id) for user_id in allowed_user_ids]
    
    
    @allowed_users.setter
    def allowed_users(self, allowed_users):
        if self._allow_users:
            return
        
        if allowed_users is None:
            allowed_user_ids = None
        
        elif isinstance(allowed_users, UserBase):
            allowed_user_ids = [allowed_users.id]
        
        elif isinstance(allowed_users, (list, set, tuple)):
            allowed_user_ids = None
            
            for user in allowed_users:
                if not isinstance(user, UserBase):
                    raise TypeError(
                        f'`allowed_users` can contain `{UserBase.__name__}` elements, got '
                        f'{type(user).__name__}; {user!r}; allowed_user_ids_input = {allowed_users!r}.'
                    )
                
                if allowed_user_ids is None:
                    allowed_user_ids = []
                
                allowed_user_ids.append(user.id)
        else:
            raise TypeError(
                f'`allowed_users` can be `None`, `{UserBase.__name__}`, `list`, `tuple`, `set` of '
                f'`{UserBase.__name__}`, got {type(allowed_users).__name__}; {allowed_users!r}.'
            )
        
        self._allowed_user_ids = allowed_user_ids
    
    
    @allowed_users.deleter
    def allowed_users(self):
        self._allowed_user_ids = None



__all__ = ('SoundboardSoundsEventHandler',)

from scarletio import RichAttributeErrorBaseType, TaskGroup

from ..core import KOKORO

from .soundboard_sounds_event_waiter import SoundboardSoundsEventWaiter


SOUNDBOARD_SOUNDS_TIMEOUT = 1.25


class SoundboardSoundsEventHandler(RichAttributeErrorBaseType):
    """
    Waits for soundboard sound events.
    
    Attributes
    ----------
    waiters : `dict` of (`int`, ``SoundboardSoundsEventWaiter``) items
        User chunk waiters.
    """
    __slots__ = ('waiters',)
    
    __event_name__ = 'soundboard_sounds'
    
    def __new__(cls):
        """
        Creates a new soundboard sounds event handler.
        """
        self = object.__new__(cls)
        self.waiters = {}
        return self
    
    
    async def __call__(self, client, event):
        """
        Called by the event parser. Ensures that the waiter's result for the given guild is set.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client, who received the respective dispatch event.
        event : ``SoundboardSoundsEvent``
            The received guild user chunk event.
        """
        try:
            waiter = self.waiters.pop(event.guild_id)
        except KeyError:
            pass
        else:
            waiter.future.set_result_if_pending(event)
    
    
    def __repr__(self):
        """Returns the soundboard sounds event handler's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append('waits at: ')
        repr_parts.append(repr(len(self.waiters)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    async def wait_for_events_in_guilds(self, guild_ids):
        """
        Waits for responses in the given guilds.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild_ids : `list` of `int`
            The guild identifiers to wait events for.
        
        Returns
        -------
        results : `list` of ``SoundboardSoundsEvent``
        """
        task_group = TaskGroup(KOKORO)
        
        # Start waiters and add them into the task group.
        waiters = self.waiters
        for guild_id in guild_ids:
            try:
                waiter = waiters[guild_id]
            except KeyError:
                waiter = SoundboardSoundsEventWaiter()
                waiters[guild_id] = waiter
            else:
                waiter.counter += 1
            
            task_group.add_future(waiter.future)
        
        # Wait all futures to complete with timeout
        future = task_group.wait_all()
        future.apply_timeout(SOUNDBOARD_SOUNDS_TIMEOUT)
        
        try:
            await future
        except TimeoutError:
            # Timeout all started waiters.
            for guild_id in guild_ids:
                try:
                    waiter = waiters.pop(guild_id)
                except KeyError:
                    pass
                else:
                    waiter.future.set_result_if_pending(None)
        
        except:
            # Other exception -> decrement waiter counters -> remove them if counter is 0
            for guild_id in guild_ids:
                try:
                    waiter = waiters[guild_id]
                except KeyError:
                    pass
                else:
                    counter = waiter.counter - 1
                    if counter == 0:
                        del waiters[guild_id]
                        waiter.future.set_result_if_pending(None)
                    else:
                        waiter.counter = counter
            
            raise
        
        # Collect results
        results = []
        
        for future in task_group.done:
            result = future.get_result()
            if (result is not None):
                results.append(result)
        
        return results



__all__ = ()

from sys import platform as PLATFORM
from zlib import decompressobj as create_zlib_decompressor, error as ZlibError
from warnings import warn

from scarletio import Task, copy_docs, from_json, repeat_timeout, skip_ready_cycle, sleep, to_json
from scarletio.web_common import ConnectionClosed, InvalidHandshake, URL, WebSocketProtocolError

from ...env import API_VERSION, CACHE_PRESENCE, LIBRARY_NAME

from ..activity import ACTIVITY_UNKNOWN
from ..core import KOKORO
from ..events.core import PARSERS
from ..events.handling_helpers import call_unknown_dispatch_event_event_handler
from ..exceptions import DiscordGatewayException, GATEWAY_EXCEPTION_CODE_TABLE
from ..guild.guild.constants import LARGE_GUILD_LIMIT

from .client_base import DiscordGatewayClientBase
from .constants import (
    GATEWAY_ACTION_CONNECT, GATEWAY_ACTION_KEEP_GOING, GATEWAY_ACTION_RESUME, GATEWAY_CONNECT_TIMEOUT,
    GATEWAY_OPERATION_CLIENT_DISPATCH, GATEWAY_OPERATION_CLIENT_HEARTBEAT,
    GATEWAY_OPERATION_CLIENT_HEARTBEAT_ACKNOWLEDGE, GATEWAY_OPERATION_CLIENT_HELLO, GATEWAY_OPERATION_CLIENT_IDENTIFY,
    GATEWAY_OPERATION_CLIENT_INVALIDATE_SESSION, GATEWAY_OPERATION_CLIENT_RECONNECT, GATEWAY_OPERATION_CLIENT_RESUME,
    GATEWAY_OPERATION_CLIENT_VOICE_STATE, LATENCY_DEFAULT, POLL_TIMEOUT
)
from .heartbeat import Kokoro
from .rate_limit import GatewayRateLimiter


async def _poll_compressed_message(web_socket):
    """
    Polls a compressed message from the given web socket.
    
    This function is a coroutine.
    
    Parameters
    ----------
    web_socket : ``WebSocketClient``
        The web socket to poll with.
    
    Returns
    -------
    raw_message : `bytes`
    
    Raises
    ------
    ConnectionClosed
        If the web socket connection closed.
    """
    buffer = None
    
    while True:
        data = await web_socket.receive()
        if data.endswith(b'\x00\x00\xff\xff'):
            if buffer is None:
                compressed_message = data
            else:
                buffer.append(data)
                compressed_message = b''.join(buffer)
            return compressed_message
        
        if buffer is None:
            buffer = []
        
        buffer.append(data)
        continue


class DiscordGatewayClientShard(DiscordGatewayClientBase):
    """
    Gateway of a client representing a shard.
    
    Attributes
    ----------
    _buffer : `list<bytes>`
        A buffer used to store not finished received payloads.
    _decompressor : `ZlibDecompressorType`
        Zlib decompressor used to decompress the received data.
    _operation_handlers : `dict<int, (instance, dict<str, object>) -> int>`
        Handler for each expected operation.
    _should_run : `bool`
        Whether the gateway should be running.
    client : ``Client``
        The owner client of the gateway.
    kokoro : `None | Kokoro`
        The heart of the gateway, sends beat-data at set intervals. If does not receives answer in time, restarts
        the gateway.
    rate_limit_handler : ``GatewayRateLimiter``
        The rate limit handler of the gateway.
    resume_gateway_url : `None | str`
        The new gateway url to which we should connect on resuming.
    sequence : `None`, `int`
        Last sequence number received.
    session_id : `None | str`
        Last session id received at `READY` event.
    shard_id : `int`
        The shard id of the gateway. If the respective client is not using sharding, it is set to `0` every time.
    web_socket : `None | WebSocketClient`
        The web socket client of the gateway.
    """
    __slots__ = (
        '_buffer', '_decompressor', '_operation_handlers', '_should_run', 'client', 'kokoro', 'rate_limit_handler',
        'resume_gateway_url', 'sequence', 'session_id', 'shard_id', 'web_socket',
    )
    
    def __new__(cls, client, shard_id):
        """
        Creates a client shard gateway.
        
        Parameters
        ----------
        client : ``Client``
            The owner client of the gateway.
        shard_id : `int`
            The shard id of the gateway.
        """
        operation_handlers = {
            GATEWAY_OPERATION_CLIENT_DISPATCH: cls._handle_operation_dispatch,
            GATEWAY_OPERATION_CLIENT_HELLO: cls._handle_operation_hello,
            GATEWAY_OPERATION_CLIENT_HEARTBEAT_ACKNOWLEDGE: cls._handle_operation_heartbeat_acknowledge,
            GATEWAY_OPERATION_CLIENT_HEARTBEAT: cls._handle_operation_heartbeat,
            GATEWAY_OPERATION_CLIENT_RECONNECT: cls._handle_operation_reconnect,
            GATEWAY_OPERATION_CLIENT_INVALIDATE_SESSION: cls._handle_operation_invalidate_session,
        }
        
        self = object.__new__(cls)
        self._buffer = []
        self._decompressor = None
        self._operation_handlers = operation_handlers
        self._should_run = False
        self.client = client
        self.kokoro = None
        self.rate_limit_handler = GatewayRateLimiter()
        self.resume_gateway_url = None
        self.sequence = -1
        self.session_id = None
        self.shard_id = shard_id
        self.web_socket = None
        return self
    
    
    @copy_docs(DiscordGatewayClientBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(' client = ')
        repr_parts.append(repr(self.client.full_name))
        
        repr_parts.append(', shard_id = ')
        repr_parts.append(repr(self.shard_id))
    
    
    async def run(self, waiter = None):
        """
        Keeps the gateway receiving message and processing it. If the gateway needs to be reconnected, reconnects
        itself. If connecting cannot succeed, because there is no internet returns `False`. If the `.client` is
        stopped, then returns `False`.
        
        This method is a coroutine.
        
        Parameters
        -----------
        waiter : `None | Future<bool>` = `None`, Optional
            A waiter future what is set, when the gateway finished connecting and started polling events.
            Its result is also set in case the task is finished before connection.
        
        Returns
        -------
        outcome : `bool<False>`
            Always `False`.
        
        Raises
        ------
        DiscordGatewayException
            The client tries to connect with bad or not acceptable intent or shard value.
        DiscordException
        """
        self._should_run = True
        client = self.client
        action = GATEWAY_ACTION_CONNECT
        
        try:
            while True:
                if (not self._should_run) or (not client.running):
                    return False
                
                try:
                    task = Task(KOKORO, self._connect(action == GATEWAY_ACTION_RESUME))
                    task.apply_timeout(GATEWAY_CONNECT_TIMEOUT)
                    action = await task
                    
                    if action != GATEWAY_ACTION_KEEP_GOING:
                        continue
                    
                    if (waiter is not None):
                        waiter.set_result_if_pending(True)
                        waiter = None
                    
                    action = await self._keep_polling_and_handling()
                
                except GeneratorExit:
                    self.abort()
                    raise
                
                except ConnectionClosed as exception:
                    code = exception.code
                    if code in GATEWAY_EXCEPTION_CODE_TABLE:
                        raise DiscordGatewayException(code) from exception
                    
                    if code not in (1000, 1006):
                        await sleep(1.0, KOKORO)
                
                except TimeoutError:
                    pass
                
                except (ConnectionError, InvalidHandshake, OSError, ValueError, WebSocketProtocolError):
                    await sleep(1.0, KOKORO)
                
                else:
                    continue
                
                action = GATEWAY_ACTION_CONNECT
                continue    
                
        finally:
            # set waiter if not set
            if (waiter is not None):
                waiter.set_result_if_pending(False)
                waiter = None
            
            # we are not running anymore.
            self._should_run = False
            
        return False
    
    
    @property
    @copy_docs(DiscordGatewayClientBase.latency)
    def latency(self):
        kokoro = self.kokoro
        if kokoro is None:
            latency = LATENCY_DEFAULT
        else:
            latency = kokoro.latency
        return latency
    
    
    def _cancel_self_and_get_web_socket(self):
        """
        Cancels the gateway except its web_socket. Returns the web_socket if still running instead.
        
        Returns
        -------
        web_socket : `None | WebSocketClient`
        """
        kokoro = self.kokoro
        if (kokoro is not None):
            kokoro.stop()
        
        web_socket = self.web_socket
        if web_socket is None:
            return None
        
        self.web_socket = None
        
        if web_socket.closed:
            return None
        
        return web_socket
    
    
    @copy_docs(DiscordGatewayClientBase.terminate)
    async def terminate(self):
        web_socket = self._cancel_self_and_get_web_socket()
        if (web_socket is not None):
            await web_socket.close(4000)
    
    
    @copy_docs(DiscordGatewayClientBase.close)
    async def close(self):
        web_socket = self._cancel_self_and_get_web_socket()
        if (web_socket is not None):
            await web_socket.close(1000)
    
    
    @copy_docs(DiscordGatewayClientBase.abort)
    def abort(self):
        web_socket = self._cancel_self_and_get_web_socket()
        if (web_socket is not None):
            web_socket.close_transport(True)
        
        self._should_run = False
    
    
    @copy_docs(DiscordGatewayClientBase.send_as_json)
    async def send_as_json(self, data):
        web_socket = self.web_socket
        if web_socket is None:
            return
        
        if not (await self.rate_limit_handler):
            return
        
        try:
            await web_socket.send(to_json(data))
        except ConnectionClosed:
            pass
    
    
    @copy_docs(DiscordGatewayClientBase.beat)
    async def beat(self):
        data = {
            'op': GATEWAY_OPERATION_CLIENT_HEARTBEAT,
            'd': self.sequence,
        }
        
        await self.send_as_json(data)
    
    
    # client base

    
    @copy_docs(DiscordGatewayClientBase.change_voice_state)
    async def change_voice_state(self, guild_id, channel_id, *, self_deaf = False, self_mute = False):
        if guild_id:
            guild_id = str(guild_id)
        else:
            guild_id = None
        
        if channel_id:
            channel_id = str(channel_id)
        else:
            channel_id = None
        
        data = {
            'op': GATEWAY_OPERATION_CLIENT_VOICE_STATE,
            'd': {
                'guild_id': guild_id,
                'channel_id': channel_id,
                'self_deaf': self_deaf,
                'self_mute': self_mute,
            },
        }
        
        await self.send_as_json(data)
    
    
    # connecting, message receive and processing
    
    async def _connect(self, resume):
        """
        Connects the gateway to Discord. If the connecting was successful will start it's `.kokoro` as well.
        
        Parameters
        ----------
        resume : `bool`
            Whether the gateway should try to resume the existing connection.
        
        Returns
        -------
        gateway_action : `int`
        
        Raises
        ------
        ConnectionError
        OSError
        ValueError
        ConnectionClosed
        InvalidHandshake
        WebSocketProtocolError
        DiscordException
        """
        self._create_kokoro()
        
        web_socket = self.web_socket
        if (web_socket is not None) and (not web_socket.closed):
            self.web_socket = None
            await web_socket.close(4000)
        
        if resume:
            gateway_url = self.resume_gateway_url
        else:
            gateway_url = None
        
        if gateway_url is None:
            gateway_url = await self.client.client_gateway_url()
        
        gateway_url = URL(f'{gateway_url}?encoding=json&v={API_VERSION}&compress=zlib-stream', True)
        
        self._decompressor = create_zlib_decompressor()
        
        self.web_socket = await self.client.http.connect_web_socket(gateway_url)
        self.kokoro.start()
        # skip one loop to wait for kokoro to start up.
        await skip_ready_cycle()
        
        # poll hello
        gateway_action = await self._poll_and_handle_received_operation()
        if gateway_action != GATEWAY_ACTION_KEEP_GOING:
            return gateway_action
        
        if not resume:
            await self._identify()
            return GATEWAY_ACTION_KEEP_GOING
        
        await self._resume()
        
        try:
            await self.web_socket.ensure_open()
        except ConnectionClosed:
            # web_socket got closed so let's just do a regular connect.
            self._clear_session()
            self.kokoro.stop()
            self.web_socket = None
            return GATEWAY_ACTION_CONNECT
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _keep_polling_and_handling(self):
        """
        Keeps polling and handling till decided otherwise.
        
        This function is a coroutine.
        
        Returns
        -------
        gateway_action : `int`
        
        Raises
        ------
        ConnectionClosed
            If the web_socket connection closed.
        """
        try:
            with repeat_timeout(POLL_TIMEOUT) as loop:
                for _ in loop:
                    action = await self._poll_and_handle_received_operation()
                    if action != GATEWAY_ACTION_KEEP_GOING:
                        return action
        
        except TimeoutError:
            # timeout, no internet probably.
            pass
        
        return GATEWAY_ACTION_CONNECT
    
    
    async def _poll_and_handle_received_operation(self):
        """
        Waits for sockets from Discord till it collected a full one. If it did, decompresses and processes it.
        Returns `True`, if the gateway should reconnect.
        
        This method is a coroutine.
        
        Returns
        -------
        gateway_action : `int`
        
        Raises
        ------
        ConnectionClosed
            If the web_socket connection closed.
        TimeoutError
            If the gateways' `.kokoro` is not beating, meanwhile it should.
        """
        web_socket = self.web_socket
        if web_socket is None:
            # No web_socket? Were the connection closed?
            return GATEWAY_ACTION_CONNECT
        
        try:
            raw_message = await _poll_compressed_message(web_socket)
        except ConnectionClosed as exception:
            # propagate a few kind of `ConnectionClosed` exceptions while swallow the rest for reconnection.
            if exception.code in (1000, 1006, 4004, 4010, 4011, 4013, 4014):
                raise
            
            return GATEWAY_ACTION_CONNECT
        
        try:
            decompressed_message = self._decompressor.decompress(raw_message)
        except ZlibError:
            # we need a full reset
            return GATEWAY_ACTION_CONNECT
        
        # This may raise `TimeoutError`
        return (await self._handle_received_operation(decompressed_message.decode('utf-8')))
    
    
    async def _handle_received_operation(self, message):
        """
        Processes the message sent by Discord. If the message is `DISPATCH`, ensures the specific parser for it and
        returns `False`. For every other operation code it calls ``._handle_special_operation`` and returns that's return.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : `bytes`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        # return True if we should reconnect
        message = from_json(message)
        
        sequence = message.get('s', None)
        if (sequence is not None):
            self.sequence = sequence
        
        operation = message['op']
        
        try:
            operation_handler = self._operation_handlers[operation]
        except KeyError:
            client = self.client
            Task(
                KOKORO,
                client.events.error(
                    client,
                    f'{type(self).__name__}._handle_received_operation',
                    f'Unknown operation: {operation!r}\nMessage: {message!r}'
                ),
            )
            return GATEWAY_ACTION_KEEP_GOING
        
        return (await operation_handler(self, message))
    
    
    async def _handle_operation_hello(self, message):
        """
        Handles a hello operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        kokoro = self.kokoro
        if kokoro is None:
            kokoro = Kokoro(self)
            self.kokoro = kokoro
        
        data = message.get('d', None)
        if (data is not None):
            interval = data['heartbeat_interval'] / 1000.0
            # send a heartbeat immediately
            kokoro.interval = interval
        
        await kokoro.beat_now()
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_heartbeat_acknowledge(self, message):
        """
        Handles a heartbeat acknowledge operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        kokoro = self.kokoro
        if kokoro is None:
            kokoro = Kokoro(self)
            self.kokoro = kokoro
        
        kokoro.answered()
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_heartbeat(self, message):
        """
        Handles a heartbeat operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        kokoro = self.kokoro
        if (kokoro is None) or (kokoro.runner is None):
            return GATEWAY_ACTION_CONNECT
        
        await self.beat()
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_reconnect(self, message):
        """
        Handles a reconnect operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        await self.terminate()
        return GATEWAY_ACTION_RESUME
    
    
    async def _handle_operation_invalidate_session(self, message):
        """
        Handles an invalidate session operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        if (data is not None) and isinstance(data, bool) and data:
            # Should sleep between 0 - 5 seconds, but actually dont need to sleep any, derp.
            # await sleep(5.0, KOKORO)
            await self.close()
            return GATEWAY_ACTION_RESUME
        
        self._clear_session()
        return GATEWAY_ACTION_CONNECT
    
    
    async def _handle_operation_dispatch(self, message):
        """
        Handles a dispatch operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        event = message['t']
        
        client = self.client
        try:
            parser = PARSERS[event]
        except KeyError:
            call_unknown_dispatch_event_event_handler(client, event, data)
            return GATEWAY_ACTION_KEEP_GOING
        
        if data is None:
            return GATEWAY_ACTION_KEEP_GOING
        
        try:
            if parser(client, data) is None:
                return GATEWAY_ACTION_KEEP_GOING
        except BaseException as err:
            Task(KOKORO, client.events.error(client, event, err))
            return GATEWAY_ACTION_KEEP_GOING
        
        if event == 'READY':
            self.session_id = data.get('session_id', None)
            self.resume_gateway_url = data.get('resume_gateway_url', None)
        
        # elif event == 'RESUMED':
            # pass
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    def _clear_session(self):
        """
        Clears current session data, disabling the option of resuming the connection.
        """
        self.session_id = None
        self.sequence = -1
        self.resume_gateway_url = None
    
    
    async def _identify(self):
        """
        Sends an `GATEWAY_OPERATION_CLIENT_IDENTIFY` packet to Discord.
        
        This method is a coroutine.
        """
        client = self.client
        activity = client._activity
        if activity is ACTIVITY_UNKNOWN:
            activity_data = None
        else:
            activity_data = activity.to_data(user = not client.bot)
        
        status_value = client._status.value
        
        data = {
            'op': GATEWAY_OPERATION_CLIENT_IDENTIFY,
            'd': {
                'token': client.token,
                'properties': {
                    'os': PLATFORM,
                    'browser': LIBRARY_NAME,
                    'device': LIBRARY_NAME,
                },
                'compress': True,                       # Whether we support compression | Discord default: False
                'large_threshold': LARGE_GUILD_LIMIT,   # between 50 and 250             | Discord default: 50
                'guild_subscriptions': CACHE_PRESENCE,  # optional                       | Discord default: False
                'intents': client.intents,              # Grip & Break down              | Discord Default: all-p-gu
                'v': 3,
                'presence': {
                    'status': status_value,
                    'game': activity_data,
                    'since': 0.0,
                    'afk': False,
                },
            },
        }
        
        shard_count = client.shard_count
        if shard_count:
            data['d']['shard'] = [self.shard_id, shard_count]
        
        await self.send_as_json(data)
    
    
    async def _resume(self):
        """
        Sends a `GATEWAY_OPERATION_CLIENT_RESUME` packet to Discord.
        
        This method is a coroutine.
        """
        data = {
            'op': GATEWAY_OPERATION_CLIENT_RESUME,
            'd': {
                'seq': self.sequence,
                'session_id': self.session_id,
                'token': self.client.token,
            },
        }
        
        await self.send_as_json(data)

    
    async def _send_json(self, data):
        """
        Internal function to send already converted json data.
        
        If the given gateway has no web_socket, or if it is closed, will not raise.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `str`
            The data to send.
        """
        web_socket = self.web_socket
        if web_socket is None:
            return
        
        if not (await self.rate_limit_handler):
            return
        
        try:
            await web_socket.send(data)
        except ConnectionClosed:
            pass
    
    
    def _create_kokoro(self):
        """
        Creates the gateway's kokoro.
        """
        kokoro = self.kokoro
        if kokoro is None:
            self.kokoro = Kokoro(self)
        else:
            kokoro.stop()
    
    
    @property
    def websocket(self):
        """
        Deprecated and will be removed in 2025 April. Use ``.web_socket`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.websocket` is deprecated and will be removed in 2025 April. '
                'Please use `.web_socket` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.web_socket



__all__ = ()

from ...core import GUILDS


def _add_embedded_activity_to_guild_cache(embedded_activity):
    """
    Adds the embedded activity to it's guild.
    
    Parameters
    ----------
    embedded_activity : ``EmbeddedActivity``
        The embedded activity to add to it's guild's cache.
    """
    try:
        guild = GUILDS[embedded_activity.guild_id]
    except KeyError:
        pass
    else:
        embedded_activities = guild.embedded_activities
        if (embedded_activities is None):
            embedded_activities = set()
            guild.embedded_activities = embedded_activities
        
        embedded_activities.add(embedded_activity)


def _remove_embedded_activity_from_guild_cache(embedded_activity):
    """
    Adds the embedded activity to it's guild.
    
    Parameters
    ----------
    embedded_activity : ``EmbeddedActivity``
        The embedded activity to add to it's guild's cache.
    """
    try:
        guild = GUILDS[embedded_activity.guild_id]
    except KeyError:
        pass
    else:
        embedded_activities = guild.embedded_activities
        if (embedded_activities is not None):
            try:
                embedded_activities.remove(embedded_activity)
            except KeyError:
                pass
            else:
                if not embedded_activities:
                    guild.embedded_activities = None



__all__ = ('ChannelType', )

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase

from ..channel_metadata import (
    ChannelMetadataBase, ChannelMetadataGuildAnnouncements, ChannelMetadataGuildCategory, ChannelMetadataGuildDirectory,
    ChannelMetadataGuildForum, ChannelMetadataGuildMedia, ChannelMetadataGuildStage, ChannelMetadataGuildStore,
    ChannelMetadataGuildText, ChannelMetadataGuildThreadAnnouncements, ChannelMetadataGuildThreadPrivate,
    ChannelMetadataGuildThreadPublic, ChannelMetadataGuildVoice, ChannelMetadataPrivate, ChannelMetadataPrivateGroup
)


from .flags import ChannelTypeFlag


@export
class ChannelType(PreinstancedBase, value_type = int):
    """
    Represents a channel's type.
    
    Attributes
    ----------
    flags : ``ChannelTypeFlag``
        The flags of the channel type defining the channel type's specifications.
    
    name : `str`
        The channel type's name.
    
    metadata_type : `type<ChannelMetadataBase>`
        The respective metadata type of the channel.
    
    value : `int`
        The identifier of the channel type.
    
    Type Attributes
    ---------------
    Every predefined channel type can be accessed as type attribute as well:
    
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | Type attribute name           | Name                          | Value | Metadata type                                 |
    +===============================+===============================+=======+===============================================+
    | unknown                       | unknown                       | -1    | ``ChannelMetadataBase``                       |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_text                    | guild text                    | 0     | ``ChannelMetadataGuildText``                  |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | private                       | private                       | 1     | ``ChannelMetadataPrivate``                    |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_voice                   | guild voice                   | 2     | ``ChannelMetadataGuildVoice``                 |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | private_group                 | private group                 | 3     | ``ChannelMetadataPrivateGroup``               |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_category                | guild category                | 4     | ``ChannelMetadataGuildCategory``              |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_announcements           | guild announcements           | 5     | ``ChannelMetadataGuildAnnouncements``         |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_store                   | guild store                   | 6     | ``ChannelMetadataGuildStore``                 |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | thread                        | thread                        | 9     | ``ChannelMetadataBase``                       |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_thread_announcements    | guild thread announcements    | 10    | ``ChannelMetadataGuildThreadAnnouncements``   |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_thread_public           | guild thread public           | 11    | ``ChannelMetadataGuildThreadPublic``          |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_thread_private          | guild thread private          | 12    | ``ChannelMetadataGuildThreadPrivate``         |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_stage                   | guild stage                   | 13    | ``ChannelMetadataGuildStage``                 |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_directory               | guild directory               | 14    | ``ChannelMetadataGuildDirectory``             |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_forum                   | guild forum                   | 15    | ``ChannelMetadataGuildForum``                 |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    | guild_media                   | guild media                   | 16    | ``ChannelMetadataGuildMedia``                 |
    +-------------------------------+-------------------------------+-------+-----------------------------------------------+
    """
    __slots__ = ('flags', 'metadata_type',)
    
    def __new__(cls, value, name = None, metadata_type = None, flags = ChannelTypeFlag.all):
        """
        Creates a new channel type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the channel type.
        
        name : `None | str` = `None`, Optional
            The default name of the channel type.
        
        metadata_type : `None | type<ChannelMetadataBase>` = `None`, Optional
            The channel type's respective metadata type.
        
        flags : ``ChannelTypeFlags`` = `ChannelTypeFlag.all`, Optional
            The flags of the channel type defining the channel type's specifications.
        """
        if metadata_type is None:
            metadata_type = ChannelMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        self.flags = flags
        return self
    
    
    unknown = P(
        -1,
        'unknown',
        ChannelMetadataBase,
        ChannelTypeFlag.all,
    )
    
    guild_text = P(
        0,
        'guild text',
        ChannelMetadataGuildText,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = True,
            invitable = True,
            private = False,
            readable = True,
            textual = True,
            thread = False,
            threadable = True,
            forum = False,
        ),
    )
    
    private = P(
        1,
        'private',
        ChannelMetadataPrivate,
        ChannelTypeFlag().update_by_keys(
            connectable = True,
            guild = False,
            guild_sortable = False,
            guild_system = False,
            invitable = False,
            private = True,
            readable = True,
            textual = True,
            thread = False,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_voice = P(
        2,
        'guild voice',
        ChannelMetadataGuildVoice,
        ChannelTypeFlag().update_by_keys(
            connectable = True,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = True,
            private = False,
            readable = True,
            textual = True,
            thread = False,
            threadable = False,
            forum = False,
        ),
    )
    
    private_group = P(
        3,
        'private group',
        ChannelMetadataPrivateGroup,
        ChannelTypeFlag().update_by_keys(
            connectable = True,
            guild = False,
            guild_sortable = False,
            guild_system = False,
            invitable = True,
            private = True,
            readable = True,
            textual = True,
            thread = False,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_category = P(
        4,
        'guild category',
        ChannelMetadataGuildCategory,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = False,
            private = False,
            readable = False,
            textual = False,
            thread = False,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_announcements = P(
        5,
        'guild announcements',
        ChannelMetadataGuildAnnouncements,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = True,
            invitable = True,
            private = False,
            readable = True,
            textual = True,
            thread = False,
            threadable = True,
            forum = False,
        ),
    )
    
    guild_store = P(
        6,
        'guild store',
        ChannelMetadataGuildStore,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = True,
            private = False,
            readable = False,
            textual = False,
            thread = False,
            threadable = False,
            forum = False,
        )
    )
    
    thread = P(
        9,
        'thread',
        ChannelMetadataBase,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = False,
            guild_sortable = False,
            guild_system = False,
            invitable = False,
            private = False,
            readable = True,
            textual = True,
            thread = True,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_thread_announcements = P(
        10,
        'guild thread announcements',
        ChannelMetadataGuildThreadAnnouncements,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = False,
            guild_system = False,
            invitable = False,
            private = False,
            readable = True,
            textual = True,
            thread = True,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_thread_public = P(
        11,
        'guild thread public',
        ChannelMetadataGuildThreadPublic,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = False,
            guild_system = False,
            invitable = False,
            private = False,
            readable = True,
            textual = True,
            thread = True,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_thread_private = P(
        12,
        'guild thread private',
        ChannelMetadataGuildThreadPrivate,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = False,
            guild_system = False,
            invitable = False,
            private = False,
            readable = True,
            textual = True,
            thread = True,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_stage = P(
        13,
        'guild stage',
        ChannelMetadataGuildStage,
        ChannelTypeFlag().update_by_keys(
            connectable = True,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = True,
            private = False,
            readable = False,
            textual = True,
            thread = False,
            threadable = False,
            forum = False,
        )
    )
    
    guild_directory = P(
        14,
        'guild directory',
        ChannelMetadataGuildDirectory,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = True,
            private = False,
            readable = True,
            textual = False,
            thread = False,
            threadable = False,
            forum = False,
        ),
    )
    
    guild_forum = P(
        15,
        'guild forum',
        ChannelMetadataGuildForum,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = True,
            private = False,
            readable = True,
            textual = False,
            thread = False,
            threadable = True,
            forum = True,
        ),
    )

    guild_media = P(
        16,
        'guild media',
        ChannelMetadataGuildMedia,
        ChannelTypeFlag().update_by_keys(
            connectable = False,
            guild = True,
            guild_sortable = True,
            guild_system = False,
            invitable = True,
            private = False,
            readable = True,
            textual = False,
            thread = False,
            threadable = True,
            forum = True,
        ),
    )



__all__ = ()

from scarletio import Compound

from ...audit_logs import AuditLog, AuditLogEntryType, AuditLogIterator
from ...bases import maybe_snowflake
from ...channel import VoiceRegion
from ...guild import (
    Guild, GuildPreview, GuildWidget, VerificationScreen, WelcomeScreen, create_partial_guild_from_data
)
from ...guild.guild.utils import GUILD_FIELD_CONVERTERS, create_new_guild_data
from ...guild.guild_incidents.utils import GUILD_INCIDENTS_FIELD_CONVERTERS
from ...guild.guild_inventory_settings.utils import GUILD_INVENTORY_SETTINGS_FIELD_CONVERTERS
from ...guild.verification_screen.utils import VERIFICATION_SCREEN_FIELD_CONVERTERS
from ...guild.welcome_screen.utils import WELCOME_SCREEN_FIELD_CONVERTERS
from ...http import DiscordApiClient
from ...payload_building import build_edit_payload
from ...role import Role
from ...onboarding import OnboardingScreen
from ...onboarding.onboarding_screen.utils import (
    ONBOARDING_FIELD_CONVERTERS, flatten_emoji_data_in_onboarding_screen_prompt_options,
    populate_prompt_ids_in_onboarding_screen_prompt_options
)
from ...user import ClientUserBase, GuildProfile, PremiumType, User, UserFlag
from ...utils import log_time_converter

from ..request_helpers import get_guild_and_id, get_guild_id, get_role_id, get_user_id


def _assert__guild_create__limit(client):
    """
    Asserts in how much guilds the client is in.
    
    Parameters
    ----------
    client : ``Client``
        The client to assert its guild creation limit of.
    """
    if client.bot:
        guild_create_limit = 10
    elif client.flags.staff:
        guild_create_limit = 200
    elif (client.premium_type is PremiumType.nitro):
        guild_create_limit = 200
    else:
        guild_create_limit = 100
    
    guild_count = len(client.guilds)
    if guild_count >= guild_create_limit:
        raise AssertionError(
            f'Guild count over creation limit; limit = {guild_create_limit}; count = {guild_count!r}.'
        )
    
    return True


class ClientCompoundGuildEndpoints(Compound):
    
    flags : UserFlag
    guild_profiles : dict
    guilds : set
    api : DiscordApiClient
    id : int
    bot : bool
    premium_type : PremiumType
    
    
    async def guild_preview_get(self, guild):
        """
        Requests the preview of a public guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The id of the guild, what's preview will be requested
        
        Returns
        -------
        preview : ``GuildPreview``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        guild_preview_data = await self.api.guild_preview_get(guild_id)
        return GuildPreview.from_data(guild_preview_data)
    
    
    async def guild_user_delete(self, guild, user, *, reason = None):
        """
        Removes the given user from the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild from where the user will be removed.
        user : ``ClientUserBase``, `int`
            The user to delete from the guild.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `user` was not given neither as ``ClientUserBase``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        user_id = get_user_id(user)
        
        await self.api.guild_user_delete(guild_id, user_id, reason)
    
    
    async def welcome_screen_get(self, guild):
        """
        Requests the given guild's welcome screen.
        
        You need to have `manage guild` permission to request the welcome screen if the guild has it disabled.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild what's welcome screen will be requested.
        
        Returns
        -------
        welcome_screen : ``WelcomeScreen``
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        welcome_screen_data = await self.api.welcome_screen_get(guild_id)
        return WelcomeScreen.from_data(welcome_screen_data)
    
    
    async def welcome_screen_edit(self, guild, welcome_screen_template = None, **keyword_parameters):
        """
        Edits the given guild's welcome screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's welcome screen will be edited.
        
        welcome_screen_template : `None`, ``WelcomeScreen``` = `None`, Optional
            Welcome screen to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the welcome screen with.
        
        Other Parameters
        ----------------
        description : `None`, `str`, Optional (Keyword only)
            Description, of what is the server about.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the welcome screen should be enabled.
        
        welcome_channels : `None`, `iterable` of ``WelcomeScreenChannel``, Optional (Keyword only)
            The featured channels by the welcome screen.
        
        Returns
        -------
        welcome_screen : ``WelcomeScreen``
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = build_edit_payload(None, welcome_screen_template, WELCOME_SCREEN_FIELD_CONVERTERS, keyword_parameters)
        welcome_screen_data = await self.api.welcome_screen_edit(guild_id, data)
        return WelcomeScreen.from_data(welcome_screen_data)
    
    
    async def verification_screen_get(self, guild):
        """
        Requests the given guild's verification screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's verification screen will be requested.

        Returns
        -------
        verification_screen : `None`, ``VerificationScreen``
        
        Raises
        ------
        TypeError
            - If `guild` was not ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the guild has no verification screen enabled, will not do any request.
        """
        guild_id = get_guild_id(guild)
        
        verification_screen_data = await self.api.verification_screen_get(guild_id)
        return VerificationScreen.from_data(verification_screen_data)
    
    
    async def verification_screen_edit(self, guild, verification_screen_template = None, **keyword_parameters):
        """
        Requests the given guild's verification screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild what's verification screen will be edited.
        
        verification_screen_template : `None`, ``VerificationScreen``` = `None`, Optional
            Verification screen to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the verification screen with.
        
        
        Other Parameters
        ----------------
        description  : `None`, `str`, Optional (Keyword only)
            The guild's description shown in the verification screen.
        
        edited_at : `None`, `datetime`, Optional (Keyword only)
            When the last version of the screen was created.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the verification screen should be enabled.
        
        steps : `None`, `tuple` of ``VerificationScreenStep``, Optional (Keyword only)
            The step in the verification screen.
        
        Returns
        -------
        verification_screen : ``VerificationScreen``
        
        Raises
        ------
        TypeError
            - If `guild` was not ``Guild``, `int`.
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = build_edit_payload(
            None, verification_screen_template, VERIFICATION_SCREEN_FIELD_CONVERTERS, keyword_parameters
        )
        verification_screen_data = await self.api.verification_screen_edit(guild_id, data)
        
        return VerificationScreen.from_data(verification_screen_data)
    
    
    async def onboarding_screen_get(self, guild):
        """
        Requests the given guild's onboarding screen.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild what's onboarding screen will be requested.
        
        Returns
        -------
        onboarding_screen : ``OnboardingScreen``
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        onboarding_screen_data = await self.api.onboarding_screen_get(guild_id)
        return OnboardingScreen.from_data(onboarding_screen_data)
    
    
    async def onboarding_screen_edit(
        self, guild, onboarding_screen_template = None, *, reason = None, **keyword_parameters
    ):
        """
        Edits the guild's onboarding screen.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to edit its onboarding screen.
        
        onboarding_screen_template : `None`, ``OnboardingScreen`` = `None`, Optional
            Onboarding screen to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the onboarding screen with.
        
        Returns
        -------
        onboarding_screen : ``OnboardingScreen``
        
        Other Parameters
        ----------------
        default_channel_ids : `None`, `iterable` of (`int`, ``Channel``), Optional (Keyword only)
            The channels' identifiers that new members get opted into automatically.
        
        enabled : `bool`, Optional (Keyword only)
            Whether onboarding is enabled.
        
        mode : ``OnboardingMode``, `int`, Optional (Keyword only)
            Onboarding mode.
        
        prompts : `None`, `iterable` of ``OnboardingPrompt``, Optional (Keyword only)
            The prompts shown during onboarding and in customize community.
        
        Raises
        ------
        TypeError
            - If `guild` is not ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        data = build_edit_payload(None, onboarding_screen_template, ONBOARDING_FIELD_CONVERTERS, keyword_parameters)
        
        # https://github.com/discord/discord-api-docs/pull/6479
        data = flatten_emoji_data_in_onboarding_screen_prompt_options(data)
        
        # this one has no ticket yet lmeow
        data = populate_prompt_ids_in_onboarding_screen_prompt_options(data, onboarding_screen_template)
        
        onboarding_screen_data = await self.api.onboarding_screen_edit(guild_id, data, reason)
        return OnboardingScreen.from_data(onboarding_screen_data)
        
    
    async def guild_get(self, guild):
        """
        Gets or updates the guild.
        
        > The client must be in the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to request.
        
        Returns
        -------
        guild : ``Guild``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        data = await self.api.guild_get(guild_id, {'with_counts': True})
        
        if guild is None:
            channel_datas = await self.api.guild_channel_get_all(guild_id)
            data['channels'] = channel_datas
            user_data = await self.api.guild_user_get(guild_id, self.id)
            data['members'] = [user_data]
            guild = Guild.from_data(data, self)
        else:
            guild._update_generic(data)
        
        return guild
    
    
    async def guild_sync(self, guild):
        """
        Syncs a guild by it's id with the wrapper. Used internally if de-sync is detected when parsing dispatch events.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to sync.

        Returns
        -------
        guild : ``Guild``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # sadly guild_get does not returns channel and voice state data at least we can request the channels
        guild, guild_id = get_guild_and_id(guild)
        
        if guild is None:
            data = await self.api.guild_get(guild_id, None)
            channel_datas = await self.api.guild_channel_get_all(guild_id)
            data['channels'] = channel_datas
            user_data = await self.api.guild_user_get(guild_id, self.id)
            data['members'] = [user_data]
            guild = Guild.from_data(data, self)
        else:
            data = await self.api.guild_get(guild_id, None)
            guild._update_generic(data)
            channel_datas = await self.api.guild_channel_get_all(guild_id)
            guild._update_channels(channel_datas)
            
            user_data = await self.api.guild_user_get(guild_id, self.id)
            try:
                profile = self.guild_profiles[guild.id]
            except KeyError:
                self.guild_profiles[guild.id] = GuildProfile.from_data(user_data)
                if self not in guild.clients:
                    guild.clients.append(self)
            else:
                profile._update_attributes(user_data)
        
        return guild

##    # Disable user syncing, takes too much time
##    async def _guild_sync_post_process(self, guild):
##        for client in CLIENTS.values():
##            try:
##                user_data = await self.api.guild_user_get(guild.id, client.id)
##           except (DiscordException, ConnectionError):
##                continue
##            try:
##                profile = client.guild_profiles[guild.id]
##            except KeyError:
##                client.guild_profiles[guild.id] = GuildProfile.from_data(user_data)
##                if client not in guild.clients:
##                    guild.clients.append(client)
##            else:
##                profile._update_attributes(user_data)
##
##        if not CACHE_USER:
##            return
##
##        old_ids = set(guild.users)
##        data = {'limit': 1000, 'after': '0'}
##        while True:
##            user_datas = await self.api.guild_users(guild.id, data)
##            for user_data in user_datas:
##                user = User._create_and_update(user_data, guild)
##                try:
##                    old_ids.remove(user.id)
##                except KeyError:
##                    pass
##
##             if len(user_datas)< 1 000:
##                 break
##
##             data['after'] = user_datas[999]['user']['id']
##
##        del data
##
##        for id_ in old_ids:
##            try:
##               user = guild.users.pop(id_)
##           except KeyError:
##               continue #huh?
##           try:
##               del user.guild_profiles[guild.id]
##           except KeyError:
##               pass # huh??


    async def guild_delete(self, guild):
        """
        Deletes the given guild. The client must be the owner of the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to delete.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        await self.api.guild_delete(guild_id)
    
    
    async def guild_create(self, name, **keyword_parameters):
        """
        Creates a guild with the given parameter. Bot accounts can create guilds only when they have less than 10.
        User account guild limit is 100, meanwhile staff guild limit is 200.
        
        This method is a coroutine.
        
        Parameters
        ----------
        name : `str`
            The name of the new guild.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Additional parameters to create the guild with.
        
        Other Parameters
        ----------------
        afk_channel_id : `None`, `int`, Optional (Keyword only)
            The id of the guild's afk channel. The id should be one of the channel's id from `channels`.
        
        afk_timeout : `None`, `int` = `None`, Optional (Keyword only)
            The afk timeout for the users at the guild's afk channel.
        
        channels : `None`, `list` of (`dict<str, object>`, ``Channel``), Optional (Keyword only)
            A list of channels of the new guild. It should contain channel data objects.
        
        explicit_content_filter_level : ``ExplicitContentFilterLevel``, `int`, Optional (Keyword only)
            The content filter level of the guild.
        
        icon : `None`, `bytes-like`, Optional (Keyword only)
            The icon of the new guild.
        
        roles : `None`, `list` of (`dict<str, object>`, ``Role``), Optional (Keyword only)
            A list of roles of the new guild. It should contain role data objects.
        
        default_message_notification_level : ``MessageNotificationLevel``, `int`, Optional (Keyword only)
            The default message notification level of the new guild.
        
        system_channel_flags : ``SystemChannelFlag``, `int`, Optional (Keyword only)
            Describe which type of messages are sent automatically to the system channel.
        
        system_channel_id: `int`, Optional (Keyword only)
            The id of the guild's system channel. The id should be one of the channel's id from `channels`.
        
        verification_level : ``VerificationLevel``, `int`, Optional (Keyword only)
            The verification level of the new guild.
        
        Returns
        -------
        guild : ``Guild``
            A partial guild made from the received data.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        assert _assert__guild_create__limit(self)
        
        data = create_new_guild_data(name = name, **keyword_parameters)
        
        data = await self.api.guild_create(data)
        # we can create only partial, because the guild data is not completed usually
        return create_partial_guild_from_data(data)
    
    
    async def guild_prune(self, guild, days, *, roles = None, count = False, reason = None):
        """
        Kicks the members of the guild which were inactive since x days.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            Where the pruning will be executed.
        days : `int`
            The amount of days since at least the users need to inactive. Can be in range [1:30].
        roles : `None`, `list` of (``Role``, `int`) = `None`, Optional (Keyword only)
            By default pruning will kick only the users without any roles, but it can defined which roles to include.
        count : `bool` = `False`, Optional (Keyword only)
            Whether the method should return how much user were pruned, but if the guild is large it will be set to
            `False` anyways. Defaults to `False`.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the guild's audit logs.
        
        Returns
        -------
        count : `None`, `int`
            The number of pruned users or `None` if `count` is set to `False`.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `roles` contain not ``Role``, neither `int` element.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `roles` was given neither as `None`, `list`.
            - If `count` was not given as `bool`.
            - If `days` was not given as `int`.
            - If `days` is out of range [1:30].
        
        See Also
        --------
        ``.guild_prune_estimate`` : Returns how much user would be pruned if ``.guild_prune`` would be called.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        if __debug__:
            if not isinstance(count, bool):
                raise AssertionError(
                    f'`count` can be `bool`, got {count.__class__.__name__}; {count!r}.'
                )
        
        if count and (guild is not None) and guild.large:
            count = False
        
        if __debug__:
            if not isinstance(days, int):
                raise AssertionError(
                    f'``days` can be `int`, got {days.__class__.__name__}; {days!r}.'
                )
            
            if days < 1 or days > 30:
                raise AssertionError(
                    f'`days can be in range [1:30], got {days!r}.'
                )
        
        data = {
            'days': days,
            'compute_prune_count': count,
        }
        
        if (roles is not None):
            if __debug__:
                if not isinstance(roles, list):
                    raise AssertionError(
                        f'`roles` can be `None`m `list` of (`{Role.__name__}`, `int`)'
                        f', got {roles.__class__.__name__}; {roles!r}.')
            
            role_ids = set()
            for index, role in enumerate(roles):
                if isinstance(role, Role):
                    role_id = role.id
                else:
                    role_id = maybe_snowflake(role)
                    if role_id is None:
                        raise TypeError(
                            f'`roles[{index}]` is not `{Role.__name__}`, `int`,'
                            f' got {role.__class__.__name__}; {role!r}; roles={roles!r}.'
                        )
                
                role_ids.add(role_id)
            
            if role_ids:
                data['include_roles'] = role_ids
        
        data = await self.api.guild_prune(guild_id, data, reason)
        return data.get('pruned', None)
    
    
    async def guild_prune_estimate(self, guild, days, *, roles = None):
        """
        Returns the amount users, who would been pruned, if ``.guild_prune`` would be called.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`.
            Where the counting of prunable users will be done.
        days : `int`
            The amount of days since at least the users need to inactive. Can be in range [1:30].
        roles : `None`, `list` of (``Role``, `int`) = `None`, Optional (Keyword only)
            By default pruning would kick only the users without any roles, but it can be defined which roles to
            include.
        
        Returns
        -------
        count : `int`
            The amount of users who would be pruned.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `roles` contain not ``Role``, neither `int` element.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `roles` was given neither as `None`, `list`.
            - If `days` was not given as `int`.
            - If `days` is out of range [1:30].
        """
        guild_id = get_guild_id(guild)
        
        if __debug__:
            if not isinstance(days, int):
                raise AssertionError(f'`days can be `int`, got {days.__class__.__name__}.')
            
            if days < 1 or days > 30:
                raise AssertionError(f'`days can be in range [1:30], got {days!r}.')
        
        data = {
            'days': days,
        }
        
        if (roles is not None):
            if __debug__:
                if not isinstance(roles, list):
                    raise AssertionError(
                        f'`roles` can be `None`, `list` of (`{Role.__name__}`, `int`)'
                        f', got {roles.__class__.__name__}; {roles!r}.'
                    )
            
            role_ids = set()
            for role in roles:
                role_id = get_role_id(role)
                role_ids.add(role_id)
            
            if role_ids:
                data['include_roles'] = role_ids
        
        data = await self.api.guild_prune_estimate(guild_id, data)
        return data.get('pruned', None)
    
    
    async def guild_edit(
        self,
        guild,
        guild_template = None,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Edits the guild with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to edit.
        
        guild_template : `None`, ``Guild`` = `None`, Optional
            Guild entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the guild with.
        
        Other Parameters
        ----------------
        afk_channel_id : `int`, ``Channel``, Optional (Keyword only)
            The afk channel or its identifier.
        
        afk_timeout : `int`, Optional (Keyword only)
            The afk timeout at the `afk_channel`.
        
        banner : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The guild's banner.
        
        boost_progress_bar_enabled : `bool`, Optional (Keyword only)
            Whether the guild has the boost progress bar enabled.
        
        explicit_content_filter_level : ``ExplicitContentFilterLevel``, `int`, Optional (Keyword only)
            The explicit content filter level of the guild.
        
        description : `None`, `str`
            Description of the guild. The guild must be a Community guild.
        
        discovery_splash : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The guild's discovery splash.
        
        features : `None`, `iterable` of `(`int`, `GuildFeature``), Optional (Keyword only)
            The guild's features.
        
        hub_type : ``HubType``, `int`, Optional (Keyword only)
            The guild's hub type.
        
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The guild's icon.
        
        invite_splash : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The guild's invite splash.
        
        locale : ``Locale``, `int`, Optional (Keyword only)
            The preferred language of the guild.
        
        default_message_notification_level : ``MessageNotificationLevel``, `int`, Optional (Keyword only)
            The default message notification level of the guild.
        
        mfa_level : ``MfaLevel``, `int`, Optional (Keyword only)
            The required multi-factor authentication level for the guild.
        
        name : `str`, Optional (Keyword only)
            The guild's name.
        
        nsfw_level : ``NsfwLevel``, `int`, Optional (Keyword only)
            The nsfw level of the guild.
        
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The guild's owner or their id.
        
        public_updates_channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's identifier where the guild's public updates should go.
        
        rules_channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where the rules of a public guild's should be.
        
        safety_alerts_channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where safety alerts are sent by Discord.
        
        system_channel_flags : ``SystemChannelFlag``, `int`, Optional (Keyword only)
            Describe which type of messages are sent automatically to the system channel.
        
        system_channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where the system messages are sent.
        
        vanity_code : `None`, `str`, Optional (Keyword only)
            The guild's vanity invite's code if it has.
        
        verification_level : ``VerificationLevel``, `int`, Optional (Keyword only)
            The minimal verification needed to join or to interact with guild.
        
        widget_channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier for which the guild's widget is for.
        
        widget_enabled : `bool`, Optional (Keyword only)
            Whether the guild's widget is enabled.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        data = build_edit_payload(guild, guild_template, GUILD_FIELD_CONVERTERS, keyword_parameters)
        
        if data:
            await self.api.guild_edit(guild_id, data, reason)
    
    
    async def guild_widget_get(self, guild):
        """
        Returns the guild's widget.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild or the guild's id, what's widget will be requested.
        
        Returns
        -------
        guild_widget : `None`, ``GuildWidget``
            If the guild has it's widget disabled returns `None` instead.
        
        Raises
        ------
        TypeError
            If `guild` was not passed neither as ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = await self.api.guild_widget_get(guild_id)
        return GuildWidget.from_data(data)
    
    
    async def guild_user_get_all(self, guild):
        """
        Requests all the users of the guild and returns them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild what's users will be requested.
        
        Returns
        -------
        users : `list` of ``ClientUserBase`` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If user caching is allowed, these users should be already loaded if the client finished starting up.
        This method takes a long time to finish for huge guilds.
        
        When using it with user account, the client's token will be invalidated.
        """
        guild_id = get_guild_id(guild)
        
        query_parameters = {'limit': 1000, 'after': 0}
        users = []
        while True:
            guild_profile_datas = await self.api.guild_user_get_chunk(guild_id, query_parameters)
            for guild_profile_data in guild_profile_datas:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id)
                users.append(user)
            
            if len(guild_profile_datas) < 1000:
                break
            
            query_parameters['after'] = user.id
        
        return users
    
    
    async def guild_get_all(self):
        """
        Requests all the guilds of the client.
        
        This method is a coroutine.
        
        Returns
        -------
        guilds : `list` of ``Guild``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the client finished starting up, all the guilds should be already loaded.
        """
        result = []
        
        query_parameters = {'after': 0, 'with_counts': True}
        while True:
            data = await self.api.guild_get_chunk(query_parameters)
            for guild_data in data:
                guild = create_partial_guild_from_data(guild_data)
                guild._update_counts_only(guild_data)
                result.append(guild)
            
            if len(data) < 100:
                break
            
            query_parameters['after'] = result[-1].id
        
        return result
    
    
    async def guild_voice_region_get_all(self, guild):
        """
        Requests the available voice regions for the given guild and returns them and the optional ones.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's regions will be requested.
        
        Returns
        -------
        voice_regions : `list` of ``VoiceRegion`` objects
            The available voice regions for the guild.
        optimals : `list` of ``VoiceRegion`` objects
            The optimal voice regions for the guild.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = await self.api.guild_voice_region_get_all(guild_id)
        voice_regions = []
        optimals = []
        for voice_region_data in data:
            region = VoiceRegion.from_data(voice_region_data)
            voice_regions.append(region)
            if voice_region_data['optimal']:
                optimals.append(region)
        
        return voice_regions, optimals
    
    
    async def voice_region_get_all(self):
        """
        Returns all the voice regions.
        
        This method is a coroutine.
        
        Returns
        -------
        voice_regions : `list` of ``VoiceRegion`` objects
            Received voice regions.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = await self.api.voice_region_get_all()
        voice_regions = []
        for voice_region_data in data:
            region = VoiceRegion.from_data(voice_region_data)
            voice_regions.append(region)
        
        return voice_regions
    
    
    async def audit_log_get_chunk(
        self,
        guild,
        limit = 100,
        *,
        before = None,
        after = None,
        user = None,
        entry_type = None,
    ):
        """
        Request a batch of audit logs of the guild and returns them. The `after`, `around` and the `before` parameters
        are mutually exclusive and they can be `int`, or as a ``DiscordEntity`` or as a `datetime`
        object.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's audit logs will be requested.
        limit : `int` = `100`, Optional
            The amount of audit logs to request. Can be between 1 and 100. Defaults to 100.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the audit log entries wer created.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the audit log entries wer created.
        user : `None`, ``ClientUserBase``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only to those, which were created by the given user.
        entry_type : `None`, ``AuditLogEntryType``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only on the given event.
        
        Returns
        -------
        audit_log : ``AuditLog``
            A container what contains the ``AuditLogEntry``-s.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, nor as `int`.
            - If `after`, `before` was passed with an unexpected type.
            - If `user` is neither `None`, ``ClientUserBase``, `int`.
            - If `entry_type` is neither `None`, ``AuditLogEntryType``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `int`.
            - If `limit` is out of the expected range [1:100].
        """
        guild, guild_id = get_guild_and_id(guild)
        
        if __debug__:
            if not isinstance(limit, int):
                raise AssertionError(
                    f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}.'
                )
            
            if limit < 1 or limit > 100:
                raise ValueError(
                    f'`limit` out of the expected range [1:100], got {limit!r}.'
                )
        
        data = {'limit': limit}
        
        if (before is not None):
            data['before'] = log_time_converter(before)
        
        if (after is not None):
            data['after'] = log_time_converter(after)
        
        if (user is not None):
            if isinstance(user, ClientUserBase):
                user_id = user.id
            
            else:
                user_id = maybe_snowflake(user)
                if user_id is None:
                    raise TypeError(
                        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
                    )
            
            data['user_id'] = user_id
        
        if (entry_type is not None):
            if isinstance(entry_type, AuditLogEntryType):
                entry_type_value = entry_type.value
            elif isinstance(entry_type, int):
                entry_type_value = entry_type
            else:
                raise TypeError(
                    f'`entry_type` can be `None`, `{AuditLogEntryType.__name__}`, `int`, got '
                    f'{entry_type.__class__.__name__}; {entry_type!r}.'
                )
            
            data['action_type'] = entry_type_value
        
        data = await self.api.audit_log_get_chunk(guild_id, data)
        return AuditLog.from_data(data, guild_id)
    
    
    async def audit_log_iterator(self, guild, *, entry_type = None, user = None):
        """
        Returns an audit log iterator for the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's audit logs will be requested.
        entry_type : `None`, ``AuditLogEntryType``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only on the given event.
        user : `None`, ``ClientUserBase``, `int` = `None`, Optional (Keyword only)
            Whether the audit logs should be filtered only to those, which were created by the given user.
        
        Returns
        -------
        audit_log_iterator : ``AuditLogIterator``
        """
        return AuditLogIterator(self, guild, entry_type = entry_type, user_id = user)
    
    
    async def guild_incidents_edit(
        self,
        guild,
        incidents_template = None,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Edits the guild's incidents with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to edit the incidents of.
        
        incidents_template : `None`, ``GuildIncidents`` = `None`, Optional
            Guild incidents to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the guild incidents with.
        
        Other Parameters
        ----------------
        direct_messages_disabled_duration : `int`, `TimeDelta`, `float`, Optional (Keyword only)
            The duration while the direct messages should be disabled in the guild in seconds or time delta.
        
        direct_messages_disabled_until : `None`, `DateTime`, Optional (Keyword only)
            Until when are the direct messages disabled in the guild.
        
        direct_messages_disabled_duration : `int`, `TimeDelta`, `float`, Optional (Keyword only)
            The duration while the invites should be disabled in the guild in seconds or time delta.
        
        invites_disabled_until : `None`, `DateTime`, Optional (Keyword only)
           Until when are the invites disabled of the guild.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        if guild is None:
            incidents = None
        else:
            incidents = guild.incidents
        
        data = build_edit_payload(
            incidents,
            incidents_template,
            GUILD_INCIDENTS_FIELD_CONVERTERS,
            keyword_parameters,
        )
        if data:
            await self.api.guild_incidents_edit(guild_id, data, reason)


    async def guild_inventory_settings_edit(
        self,
        guild,
        inventory_settings_template = None,
        *,
        reason = None,
        **keyword_parameters,
    ):
        """
        Edits the guild's inventory settings with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to edit the inventory settings of.
        
        inventory_settings_template : `None`, ``GuildInventorySettings`` = `None`, Optional
            Guild inventory settings to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to edit the guild inventory settings with.
        
        Other Parameters
        ----------------
        emoji_pack_collectible : `bool`, Optional (Keyword only)
            Whether you can collect this guild's emojis and use it across all guilds.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        if guild is None:
            inventory_settings = None
        else:
            inventory_settings = guild.inventory_settings
        
        data = build_edit_payload(
            inventory_settings,
            inventory_settings_template,
            GUILD_INVENTORY_SETTINGS_FIELD_CONVERTERS,
            keyword_parameters,
        )
        if data:
            await self.api.guild_inventory_settings_edit(guild_id, data, reason)



__all__ = ('register',)

from functools import partial as partial_func

from .command import Command, CommandCategory, CommandFunction
from .command.helpers import get_function_description, get_function_name


def register(func = ..., *, available = True, aliases = None, description = None, into = None, name = None):
    """
    Registers the given command.
    
    Parameters
    ----------
    func : `None`, `FunctionType`
        The function to register.
    aliases : `None`, `str`, `iterable` of `str` = `None`, Optional (Keyword only)
        Alternative names for the command.
    available : `bool` = `True`, Optional (Keyword only)
        Whether the command is available.
    description : `None`, `str` = `None`, Optional (Keyword only)
        Alternative description to use instead of the function's.
    into : `None`, ``Command``, ``CommandCategory``, ``CommandFunction`` = `None`, Optional (Keyword only)
        The command to register currently added one into.
    name : `None`, `str` = `None`, Optional (Keyword only)
        Alternative name to use instead of the function's.
    
    Returns
    -------
    command / decorator : ``Command``, ``CommandFunction`` / `functools.partial`
        The registered command or a partial function to add it.
    
    Raises
    ------
    TypeError
        - Invalid parameter type.
    RuntimeError
        - `into`'s parent command already garbage collected.
    """
    if func is ...:
        return partial_func(
            register, aliases = aliases, available = available, description = description, into = into, name = name
        )
    
    name = get_function_name(func, name)
    description = get_function_description(func, description)
    
    if (into is None):
        command = Command(name, aliases, available)
    
    elif isinstance(into, Command):
        command = into
    
    elif isinstance(into, CommandCategory):
        command = into
    
    elif isinstance(into, CommandFunction):
        command = into.get_parent()
        if (command is None):
            raise RuntimeError(
                f'`into`\'s parent already garbage collected, got {into!r}.'
            )
    else:
        raise TypeError(
            f'`into` can be `None`, `{Command.__name__}`, `{CommandCategory.__name__}`, `{CommandFunction.__name__}`,'
            f' got {into.__class__.__name__}; {into!r}.'
        )
    
    if (func is not None):
        command = command.register_command_function(func, name, description)
    elif (into is not None):
        command = command.register_command_category(name)
    
    return command



__all__ = ('Stats', )

from scarletio import RichAttributeErrorBaseType

from .constants import (
    LAVALINK_KEY_STATS_CPU, LAVALINK_KEY_STATS_CPU_CORES, LAVALINK_KEY_STATS_CPU_LAVALINK_LOAD,
    LAVALINK_KEY_STATS_CPU_SYSTEM_LOAD, LAVALINK_KEY_STATS_FRAME, LAVALINK_KEY_STATS_FRAME_DEFICIT,
    LAVALINK_KEY_STATS_FRAME_NULLED, LAVALINK_KEY_STATS_FRAME_SENT, LAVALINK_KEY_STATS_MEMORY,
    LAVALINK_KEY_STATS_MEMORY_ALLOCATED, LAVALINK_KEY_STATS_MEMORY_FREE, LAVALINK_KEY_STATS_MEMORY_RESERVABLE,
    LAVALINK_KEY_STATS_MEMORY_USED, LAVALINK_KEY_STATS_PLAYER_COUNT, LAVALINK_KEY_STATS_PLAYING_PLAYER_COUNT,
    LAVALINK_KEY_STATS_UPTIME
)


class Stats(RichAttributeErrorBaseType):
    """
    Represents the stats of a Lavalink node.
    
    Attributes
    ----------
    _total_penalty : `None`, `float`
        Total resource penalty of the respective node.
    cpu_cores : `int`
        The amount of cpu cores the system of the node has.
    cpu_lavalink_load : `int`
        The CPU load generated by Lavalink.
    cpu_system_load : `int`
        The overall CPU load of the system.
    frame_sent : `int`
        The number of frames sent.
    frame_nulled : `int`
        The number of null frames.
    frame_deficit : `int`
        The number of missing frames.
    memory_allocated : `int`
        The amount of memory allocated to the node.
    memory_free : `int`
        The amount of memory free to the node.
    memory_reservable : `int`
        The amount of memory reservable to the node.
    memory_used : `int`
        The amount of memory that is used by the node.
    player_count : `int`
        The amount of players connected to the node.
    playing_player_count : `int`
        The amount of players that are playing in the node.
    uptime : `float`
        How long the node has been running for in seconds.
    """
    __slots__ = (
        '_total_penalty', 'cpu_cores', 'cpu_lavalink_load', 'cpu_system_load', 'frame_deficit', 'frame_nulled',
        'frame_sent', 'memory_allocated', 'memory_free', 'memory_reservable', 'memory_used', 'player_count',
        'playing_player_count', 'uptime'
    )
    
    def __init__(self, data):
        """
        Creates a new stats instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Deserialized json.
        """
        self.uptime = data[LAVALINK_KEY_STATS_UPTIME]*1000.0
        self.player_count = data[LAVALINK_KEY_STATS_PLAYER_COUNT]
        self.playing_player_count = data[LAVALINK_KEY_STATS_PLAYING_PLAYER_COUNT]
        
        memory_data = data[LAVALINK_KEY_STATS_MEMORY]
        self.memory_allocated = memory_data[LAVALINK_KEY_STATS_MEMORY_ALLOCATED]
        self.memory_free = memory_data[LAVALINK_KEY_STATS_MEMORY_FREE]
        self.memory_reservable = memory_data[LAVALINK_KEY_STATS_MEMORY_RESERVABLE]
        self.memory_used = memory_data[LAVALINK_KEY_STATS_MEMORY_USED]
        
        cpu_data = data[LAVALINK_KEY_STATS_CPU]
        self.cpu_cores = cpu_data[LAVALINK_KEY_STATS_CPU_CORES]
        self.cpu_lavalink_load = cpu_data[LAVALINK_KEY_STATS_CPU_SYSTEM_LOAD]
        self.cpu_system_load = cpu_data[LAVALINK_KEY_STATS_CPU_LAVALINK_LOAD]
        
        try:
            frame_data = data[LAVALINK_KEY_STATS_FRAME]
        except KeyError:
            frame_sent = 0
            frame_nulled = 0
            frame_deficit = 0
        else:
            frame_sent = frame_data.get(LAVALINK_KEY_STATS_FRAME_SENT, 0)
            frame_nulled = frame_data.get(LAVALINK_KEY_STATS_FRAME_NULLED, 0)
            frame_deficit = frame_data.get(LAVALINK_KEY_STATS_FRAME_DEFICIT, 0)
        self.frame_sent = frame_sent
        self.frame_nulled = frame_nulled
        self.frame_deficit = frame_deficit
        
        self._total_penalty = None
    
    
    @property
    def total_penalty(self):
        """
        Returns the resource penalty of the node.
        
        Returns
        -------
        penalty : `float`
        """
        total_penalty = self._total_penalty
        if (total_penalty is None):
            total_penalty = self._get_total_penalty()
        
        return total_penalty
    
    
    def _get_total_penalty(self):
        """
        Returns the resource penalty of the node.
        
        Used by the ``.total_penalty`` property, if penalty was not yet set.
        
        Returns
        -------
        penalty : `float`
        """
        player_penalty = self.playing_player_count
        cpu_penalty = 1.05 ** (100 * self.cpu_system_load) * 10 - 10
        
        frame_nulled = self.frame_nulled
        if frame_nulled:
            null_frame_penalty = ((1.03 ** (frame_nulled / 6)) * 300 - 300) * 2
        else:
            null_frame_penalty = 0
        
        frame_deficit = self.frame_deficit
        if frame_deficit:
            deficit_frame_penalty = (1.03 ** ((frame_deficit / 6))) * 600 - 600
        else:
            deficit_frame_penalty = 0.0
        
        total_penalty = player_penalty + cpu_penalty + null_frame_penalty + deficit_frame_penalty
        self._total_penalty = total_penalty
        return total_penalty
    
    
    def __repr__(self):
        """Returns the stat's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(', cpu_cores = ')
        repr_parts.append(repr(self.cpu_cores))
        
        repr_parts.append(', cpu_lavalink_load = ')
        repr_parts.append(repr(self.cpu_lavalink_load))
        
        repr_parts.append(', cpu_system_load = ')
        repr_parts.append(repr(self.cpu_system_load))
        
        repr_parts.append(', frame_sent = ')
        repr_parts.append(repr(self.frame_sent))
        
        repr_parts.append(', frame_nulled = ' )
        repr_parts.append(repr(self.frame_nulled))
        
        repr_parts.append(', frame_deficit = ')
        repr_parts.append(repr(self.frame_deficit))
        
        repr_parts.append(', memory_allocated = ')
        repr_parts.append(repr(self.memory_allocated))
        
        repr_parts.append(', memory_free = ')
        repr_parts.append(repr(self.memory_free))
        
        repr_parts.append(', memory_reservable = ')
        repr_parts.append(repr(self.memory_reservable))
        
        repr_parts.append(', memory_used = ')
        repr_parts.append(repr(self.memory_used))
        
        repr_parts.append(', player_count = ')
        repr_parts.append(repr(self.player_count))
        
        repr_parts.append(', playing_player_count = ')
        repr_parts.append(repr(self.playing_player_count))
        
        repr_parts.append(', uptime = ')
        repr_parts.append(self.uptime.__format__('.3f'))
        repr_parts.append('s')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)



__all__ = ('HangType',)

from ...bases import Preinstance as P, PreinstancedBase


ACTIVITY_NAME_HANGING_DEFAULT = 'Right now, I\'m -'


def _name_getter_custom(activity):
    """
    Name getter for ``HangType`` for its custom instance.
    
    Parameters
    ----------
    activity : ``Activity``
        The activity to get its name of.
    
    Returns
    -------
    name : `str`
    """
    details = activity.details
    emoji = activity.emoji
    if (details is None):
        if (emoji is None):
            name = ACTIVITY_NAME_HANGING_DEFAULT
        else:
            name = emoji.as_emoji
    else:
        if (emoji is None):
            name = details
        else:
            name = f'{emoji} {details}'

    return name


class HangType(PreinstancedBase, value_type = str):
    """
    Represents a hanging activity' type.
    
    Attributes
    ----------
    name : `str`
        The hang type's name.
    
    name_getter : `(Activity) -> str`
        Function used when getting the hanging activity's name.
    
    value : `str`
        The Discord side identifier value of the hang type.
    
    Type Attributes
    ---------------
    Every predefined hang type can be accessed as type attribute as well:
    
    +-----------------------+---------------+---------------+
    | Type attribute name   | name          | value         |
    +=======================+===============+===============+
    | none                  | none          |               |
    +-----------------------+---------------+---------------+
    | be_right_back         | be right back | brb           |
    +-----------------------+---------------+---------------+
    | chilling              | chilling      | chilling      |
    +-----------------------+---------------+---------------+
    | custom                | custom        | custom        |
    +-----------------------+---------------+---------------+
    | focusing              | focusing      | focusing      |
    +-----------------------+---------------+---------------+
    | gaming                | gaming        | gaming        |
    +-----------------------+---------------+---------------+
    | eating                | eating        | eating        |
    +-----------------------+---------------+---------------+
    | travelling            | travelling    | in-transit    |
    +-----------------------+---------------+---------------+
    | watching              | watching      | watching      |
    +-----------------------+---------------+---------------+
    """
    NAME_GETTER_DEFAULT = lambda activity: ACTIVITY_NAME_HANGING_DEFAULT
    
    __slots__ = ('name_getter', )
    
    def __new__(cls, value, name = None, name_getter = None):
        """
        Creates a new hang type.
        
        Parameters
        ----------
        value : `str`
            The Discord side identifier value of the hang type.
        
        name : `None | str` = `None`, Optional
            The name of the hang type.
        
        name_getter : `None | FunctionType` = `None`, Optional
            The hang type's name getter.
        """
        if name is None:
            name = value.casefold().replace('-', ' ')
        
        if name_getter is None:
            name_getter = cls.NAME_GETTER_DEFAULT
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.name_getter = name_getter
        return self
    
    
    # predefined
    none = P('', 'none', NAME_GETTER_DEFAULT)
    be_right_back = P('brb', 'be right back', (lambda activity: 'Gonna BRB'))
    chilling = P('chilling', 'chilling', (lambda activity: 'Chilling'))
    custom = P('custom', 'custom', _name_getter_custom)
    eating = P('eating', 'eating', (lambda activity: 'Grubbin'))
    focusing = P('focusing', 'focusing', (lambda activity: 'In the zone'))
    gaming = P('gaming', 'gaming', (lambda activity: 'GAMING'))
    travelling = P('in-transit', 'travelling', (lambda activity: 'Wandering IRL'))
    watching = P('watching', 'watching', (lambda activity: 'Watchin\' stuff'))



__all__ = (
    'create_button', 'create_channel_select', 'create_media_gallery', 'create_mentionable_select',
    'create_role_select', 'create_row', 'create_separator', 'create_string_select', 'create_text', 'create_text_input',
    'create_user_select'
)

from .component import Component, ComponentType


def create_button(
    label = ...,
    emoji = ...,
    *,
    custom_id = ...,
    enabled = ...,
    sku_id = ...,
    style = ...,
    url = ...,
):
    """
    Creates a new button component.
    
    Parameters
    -----------
    label : `None`, `str`, Optional
        Label of the component.
    
    emoji : `None`, ``Emoji``, Optional
        Emoji of the button if applicable.
    
    custom_id : `None`, `str`, Optional (Keyword only)
        Custom identifier to detect which button was clicked by the user.
        
        > Mutually exclusive with the `url` and `sku_id` fields.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    sku_id : `int`, ``SKU``, Optional (Keyword only)
        Purchasable stock keeping unit identifier.
        
        > Mutually exclusive with the `custom_id` and `url` fields.
    
    style : `None`, ``ButtonStyle``, `int`, Optional (Keyword only)
        The button's style.
    
    url : `None`, `str`, Optional (Keyword only)
        Url to redirect to when clicking on the button.
        
        > Mutually exclusive with the `custom_id` and `sku_id` fields.
    
    Returns
    -------
    button : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
        - `url` and `custom_id` are mutually exclusive.
    """
    return Component(
        ComponentType.button,
        button_style = style,
        custom_id = custom_id,
        label = label,
        enabled = enabled,
        emoji = emoji,
        sku_id = sku_id,
        url = url,
    )


def create_row(
    *components,
):
    """
    Creates a row component from the given components.
    
    Parameters
    ----------
    *components : ``Component``
        Sub components.
    
    Returns
    -------
    row : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    """
    return Component(
        ComponentType.row,
        components = components,
    )


def create_string_select(
    options,
    custom_id = ...,
    *,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
):
    """
    Creates a new string select.
    
    Parameters
    ----------
    options : `None` or `iterable` of ``StringSelectOption``
        Options of the select.
    
    custom_id : `None`, `str`, Optional
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    string_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.string_select,
        custom_id = custom_id,
        enabled = enabled,
        options = options,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_text_input(
    label = ...,
    *,
    custom_id = ...,
    max_length = ...,
    min_length = ...,
    placeholder = ...,
    required = ...,
    style = ...,
    value = ...,
):
    """
    Creates a new text input. Can be used inside of forms.
    
    Parameters
    ----------
    label : `None`, `str`, Optional
        Label of the component.
    
    custom_id : `None`, `str`, Optional (Keyword only)
        Custom identifier to detect which text input was clicked by the user.
        
        If not given, is autogenerated form the `label` parameter.
    
    max_length : `int`, Optional (Keyword only)
        The maximal length of the inputted text.
    
    min_length : `int`, Optional (Keyword only)
        The minimal length of the inputted text.
    
    placeholder : `None`, `str`, Optional (Keyword only)
        Placeholder text of the select.
    
    required : `None`, `bool`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
        
        If not given, or given as `None`, will default to `True` if `min_length` is defined as higher than `0`.
    
    style : `None`, ``TextInputStyle``, `int`, Optional (Keyword only)
        The text input's style.
    
    value : `None`, `str`, Optional (Keyword only)
        The text input's default value.
    
    Returns
    -------
    text_input : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.text_input,
        custom_id = custom_id,
        label = label,
        max_length = max_length,
        min_length = min_length,
        placeholder = placeholder,
        required = required,
        text_input_style = style,
        value = value,
    )


def create_user_select(
    custom_id = ...,
    *,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
):
    """
    Creates a new user select.
    
    Parameters
    ----------
    custom_id : `None`, `str`, Optional
        Custom identifier to detect which component was used by the user.
    
    default_values : `None | iterable<ClientUserBase | EntitySelectDefaultValue | tuple>`, Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    user_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.user_select,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_role_select(
    custom_id = ...,
    *,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
):
    """
    Creates a new role select.
    
    Parameters
    ----------
    custom_id : `None`, `str`, Optional
        Custom identifier to detect which component was used by the user.
    
    default_values : `None | iterable<Role | EntitySelectDefaultValue | tuple>`, Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    role_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.role_select,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_mentionable_select(
    custom_id = ...,
    *,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
):
    """
    Creates a new mentionable (user & role) select.
    
    Parameters
    ----------
    custom_id : `None`, `str`, Optional
        Custom identifier to detect which component was used by the user.
    
    default_values : `None | iterable<ClientUserBase | Role | EntitySelectDefaultValue | tuple>` \
            , Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    mentionable_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.mentionable_select,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_channel_select(
    custom_id = ...,
    *,
    channel_types = ...,
    default_values = ...,
    enabled = ...,
    placeholder = ...,
    max_values = ...,
    min_values = ...,
):
    """
    Creates a new channel select.
    
    Parameters
    ----------
    custom_id : `None`, `str`, Optional
        Custom identifier to detect which component was used by the user.
    
    channel_types : `None`, `tuple` of (``ChannelType``, `int`), Optional (Keyword only)
        The allowed channel types by the select.
    
    default_values : `None | iterable<ClientUserBase | Channel | EntitySelectDefaultValue | tuple>` \
            , Optional (Keyword only)
        Entities presented in the select by default.
    
    enabled : `bool`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    channel_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.channel_select,
        channel_types = channel_types,
        custom_id = custom_id,
        default_values = default_values,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_text(
    content = ...,
):
    """
    Creates a new text component. Can be used inside of forms.
    
    Parameters
    ----------
    content : `None`, `str`, Optional
        The content shown on the component.
    
    Returns
    -------
    text : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.text,
        content = content,
    )


def create_media_gallery(
    *items,
):
    """
    Creates a new media gallery component. Can be used inside of forms.
    
    Parameters
    ----------
    *items : ``MediaItem``
        The media items shown on the component.
    
    Returns
    -------
    text : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.media_gallery,
        items = items,
    )


def create_separator(
    *,
    divider = ...,
    spacing_size = ...,
):
    """
    Creates a new separator component. Can be used inside of forms.
    
    Parameters
    ----------
    divider : `bool`, Optional (Keyword only)
        Whether the separator should contain a divider.
    spacing_size : ``SeparatorSpacingSize``, `int`, Optional (Keyword only)
        The separator's spacing's size.
    
    Returns
    -------
    separator : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.separator,
        divider = divider,
        spacing_size = spacing_size,
    )



__all__ = ()


MAX_PRESENCES_DEFAULT = 0
MAX_USERS_DEFAULT = 250000 # Most already has 500000 by default
MAX_STAGE_CHANNEL_VIDEO_USERS_DEFAULT = 50
MAX_VOICE_CHANNEL_VIDEO_USERS_DEFAULT = 25

NAME_LENGTH_MIN = 2
NAME_LENGTH_MAX = 100

AFK_TIMEOUT_DEFAULT = 0
AFK_TIMEOUT_OPTIONS = frozenset((0, 60, 300, 900, 1800, 3600),)

DESCRIPTION_LENGTH_MAX = 120

LARGE_GUILD_LIMIT = 250 # can be between 50 and 250


# State flags

GUILD_STATE_MASK_SOUNDBOARD_SOUNDS_CACHED = 1 << 0
GUILD_STATE_MASK_CACHE_BOOSTERS = 1 << 1

GUILD_STATE_MASK_CACHE_ALL = GUILD_STATE_MASK_CACHE_BOOSTERS

# Constants for events

EMOJI_EVENT_NONE = 0
EMOJI_EVENT_CREATE = 1
EMOJI_EVENT_DELETE = 2
EMOJI_EVENT_UPDATE = 3


SOUNDBOARD_SOUND_EVENT_NONE = 0
SOUNDBOARD_SOUND_EVENT_CREATE = 1
SOUNDBOARD_SOUND_EVENT_DELETE = 2
SOUNDBOARD_SOUND_EVENT_UPDATE = 3


STICKER_EVENT_NONE = 0
STICKER_EVENT_CREATE = 1
STICKER_EVENT_DELETE = 2
STICKER_EVENT_UPDATE = 3


VOICE_STATE_EVENT_NONE = 0
VOICE_STATE_EVENT_JOIN = 1
VOICE_STATE_EVENT_LEAVE = 2
VOICE_STATE_EVENT_UPDATE = 3
VOICE_STATE_EVENT_MOVE = 4



__all__ = ('ApplicationCommandOptionMetadataString',)

from scarletio import copy_docs

from .constants import MAX_LENGTH_DEFAULT, MIN_LENGTH_DEFAULT
from .fields import (
    parse_max_length, parse_min_length, put_max_length, put_min_length, validate_max_length,
    validate_min_length
)
from .primitive import ApplicationCommandOptionMetadataPrimitive


class ApplicationCommandOptionMetadataString(ApplicationCommandOptionMetadataPrimitive):
    """
    String parameter application command option metadata.
    
    Attributes
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
    
    choices : `None`, `tuple` of ``ApplicationCommandOptionChoice``
        Choices for the user to pick from.
    
    max_length : `int`
        The maximum input length allowed for this option.
    
    min_length : `int`
        The minimum input length allowed for this option.
    
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    """
    TYPE = str
    
    __slots__ = ('max_length', 'min_length')
    
    def __new__(cls, *, autocomplete = ..., choices = ..., max_length = ..., min_length = ..., required = ...):
        """
        Creates a new string application command option metadata with the given parameters.
        
        Parameters
        ----------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            Choices for the user to pick from.
        
        max_length : `int`, Optional (Keyword only)
            The maximum input length allowed for this option.
        
        min_length : `int`, Optional (Keyword only)
            The minimum input length allowed for this option.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # max_length
        if max_length is ...:
            max_length = MAX_LENGTH_DEFAULT
        else:
            max_length = validate_max_length(max_length)
        
        # min_length
        if min_length is ...:
            min_length = MIN_LENGTH_DEFAULT
        else:
            min_length = validate_min_length(min_length)
        
        # Construct
        new = ApplicationCommandOptionMetadataPrimitive.__new__(
            cls,
            autocomplete = autocomplete,
            choices = choices,
            required = required,
        )
        new.max_length = max_length
        new.min_length = min_length
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            autocomplete = keyword_parameters.pop('autocomplete', ...),
            choices = keyword_parameters.pop('choices', ...),
            max_length = keyword_parameters.pop('max_length', ...),
            min_length = keyword_parameters.pop('min_length', ...),
            required = keyword_parameters.pop('required', ...),
        )
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.from_data)
    def from_data(cls, data):
        self = super(ApplicationCommandOptionMetadataString, cls).from_data(data)
        self.max_length = parse_max_length(data)
        self.min_length = parse_min_length(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.to_data)
    def to_data(self, *, defaults = False):
        data = ApplicationCommandOptionMetadataPrimitive.to_data(self, defaults = defaults)
        put_max_length(self.max_length, data, defaults)
        put_min_length(self.min_length, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        ApplicationCommandOptionMetadataPrimitive._add_type_specific_repr_fields(self, repr_parts)
        
        repr_parts.append(', max_length = ')
        repr_parts.append(repr(self.max_length))
        
        repr_parts.append(', min_length = ')
        repr_parts.append(repr(self.min_length))
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ApplicationCommandOptionMetadataPrimitive._is_equal_same_type(self, other):
            return False
        
        if self.max_length != other.max_length:
            return False
        
        if self.min_length != other.min_length:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.__hash__)
    def __hash__(self):
        hash_value = ApplicationCommandOptionMetadataPrimitive.__hash__(self)
        
        # max_length
        hash_value ^= self.max_length << 6
        
        # min_length
        hash_value ^= self.min_length  << 10
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.copy)
    def copy(self):
        new = ApplicationCommandOptionMetadataPrimitive.copy(self)
        new.max_length = self.max_length
        
        # min_length
        new.min_length = self.min_length
        
        return new
    
    
    def copy_with(self, *, autocomplete = ..., choices = ..., max_length = ..., min_length = ..., required = ...):
        """
        Copies the string application command option metadata with the given fields.
        
        Parameters
        ----------
        autocomplete : `bool`, Optional (Keyword only)
            Whether the option supports auto completion.
        
        choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            Choices for the user to pick from.
        
        max_length : `int`, Optional (Keyword only)
            The maximum input length allowed for this option.
        
        min_length : `int`, Optional (Keyword only)
            The minimum input length allowed for this option.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # max_length
        if max_length is ...:
            max_length = self.max_length
        else:
            max_length = validate_max_length(max_length)
        
        # min_length
        if min_length is ...:
            min_length = self.min_length
        else:
            min_length = validate_min_length(min_length)
        
        # Construct
        new = ApplicationCommandOptionMetadataPrimitive.copy_with(
            self, 
            autocomplete = autocomplete,
            choices = choices,
            required = required,
        )
        new.max_length = max_length
        new.min_length = min_length
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataPrimitive.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            autocomplete = keyword_parameters.pop('autocomplete', ...),
            choices = keyword_parameters.pop('choices', ...),
            max_length = keyword_parameters.pop('max_length', ...),
            min_length = keyword_parameters.pop('min_length', ...),
            required = keyword_parameters.pop('required', ...),
        )



__all__ = ()

from ...field_parsers import entity_id_parser_factory, nullable_entity_parser_factory
from ...field_putters import (
    entity_id_optional_putter_factory, entity_id_putter_factory, nullable_entity_putter_factory
)
from ...field_validators import entity_id_validator_factory, nullable_entity_validator_factory

from ..embedded_activity_location import EmbeddedActivityLocation
from ..embedded_activity_user_state import EmbeddedActivityUserState


# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', NotImplemented, include = 'Application')


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')


# id

parse_id = entity_id_parser_factory('instance_id')
put_id = entity_id_putter_factory('instance_id')
validate_id = entity_id_validator_factory('id')


# launch_id

parse_launch_id = entity_id_parser_factory('launch_id')
put_launch_id = entity_id_putter_factory('launch_id')
validate_launch_id = entity_id_validator_factory('launch_id')


# location

parse_location = nullable_entity_parser_factory('location', EmbeddedActivityLocation)
put_location = nullable_entity_putter_factory('location', EmbeddedActivityLocation)
validate_location = nullable_entity_validator_factory('location', EmbeddedActivityLocation)


# user_states

def parse_user_states(data, user_states, guild_id = 0):
    """
    Parses the embedded activity user states from the given data.
    
    
    Parameters
    ----------
    input_data : `dict<str, object>`
        data to parse from.
    
    user_states : `dict<int, EmbeddedActivityUserState>`
        The embedded activity's current user states.
    
    guild_id : `int` = `0`, Optional
        The respective guild's identifier.
    
    Returns
    -------
    user_states : `dict<int, EmbeddedActivityUserState>`
    """
    # No need to keep old entities, we just want to use the same dictionary.
    user_states.clear()
    
    user_state_datas = data.get('participants', None)
    if (user_state_datas is not None) and user_state_datas:
        for user_state_data in user_state_datas:
            user_state = EmbeddedActivityUserState.from_data(user_state_data, guild_id)
            user_states[user_state.user_id] = user_state
    
    return user_states


def put_user_states(user_states, data, defaults, *, guild_id = 0):
    """
    Parses the embedded activity user states from the given data.
    
    
    Parameters
    ----------
    user_states : `dict<int, EmbeddedActivityUserState>`
        The user states to serialize.
    
    data : `dict<str, object>`
        Data to serialize to.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    guild_id : `int` = `0`, Optional (Keyword only)
        The respective guild's identifier.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    user_states_data = []
    
    for user_state in user_states.values():
        user_states_data.append(user_state.to_data(defaults = defaults, guild_id = guild_id))
    
    data['participants'] = user_states_data
    return data


def validate_user_states(user_states):
    """
    Validates the given user states value.
    
    Parameters
    ----------
    user_states : `None | iterable<EmbeddedActivityUserState> | dict<int, EmbeddedActivityUserState>`
        The field value to validate.
    
    Returns
    -------
    entity_dictionary : `dict<int, EmbeddedActivityUserState>`
    
    Raises
    ------
    TypeError
        - If `user_states`'s or it's elements type is incorrect.
    """
    validated = {}
    
    if user_states is None:
        return validated
    
    if isinstance(user_states, dict):
        iterator = iter(user_states.values())
    
    elif (getattr(user_states, '__iter__', None) is not None):
        iterator = iter(user_states)
    
    else:
        raise TypeError(
            f'`user_states` can be `None`, `dict` of (`int`, `{EmbeddedActivityUserState.__name__}`) items, '
            f'`iterable` of `{EmbeddedActivityUserState.__name__}` items, '
            f'got {type(user_states).__name__}; {user_states!r}.'
        )
    
    for element in iterator:
        if not isinstance(element, EmbeddedActivityUserState):
            raise TypeError(
                f'`user_states` elements can be `{EmbeddedActivityUserState.__name__}`, got {type(element).__name__}; '
                f'{element!r}; user_states = {user_states!r}.'
            )
        
        validated[element.user_id] = element
    
    return validated
